package com.ibm.btools.te.ilm.heuristics.fdl.util;
/*
* Licensed Materials - Property of IBM
* 5724-I74, 5724-I75
*  Copyright IBM Corporation 2003, 2009. All rights reserved.
* US Government Users Restricted Rights - Use, duplication or disclosure
* restricted by GSA ADP Schedule Contract with IBM Corp.
*/

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import y.base.Node;
import y.layout.BufferedLayouter;
import y.layout.LayoutGraph;
import y.layout.NodeLayout;
import y.layout.OrientationLayouter;
import y.layout.hierarchic.HierarchicLayouter;

import com.ibm.btools.cef.gef.layouts.AutoLayoutGraph;
import com.ibm.btools.cef.main.CefLiterals;
import com.ibm.btools.fdl.model.Activity;
import com.ibm.btools.fdl.model.Block;
import com.ibm.btools.fdl.model.ControlFlow;
import com.ibm.btools.fdl.model.DataFlow;
import com.ibm.btools.fdl.model.ModelFactory;
import com.ibm.btools.fdl.model.PaperSize;
import com.ibm.btools.fdl.model.ScreenPosition;
import com.ibm.btools.fdl.model.Sink;
import com.ibm.btools.fdl.model.Source;
import com.ibm.btools.te.ilm.LoggingUtil;

/**
 * The abstract LayoutManager class provides functions for applying layout algorithms
 * to a fdl graph.  The yFile graph is generated from a FDL model and layouted.
 * The generated coordinates for each node then converted to FDL x, y coordinates. 
 * 
 */

public abstract class AbstractFdlLayoutManager {
	
	public static final String copyright = "Licensed Material - Property of IBM " + " 5724-I74, 5724-I75 " + "(C) Copyright IBM Corporation 2003, 2009. All Rights Reserved. " + "U.S. Government Users Restricted Rights - Use, duplication or disclosure " + "restricted by GSA ADP Schedule Contract with IBM Corp.".intern(); //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$ //$NON-NLS-4$ //$NON-NLS-5$
	
	/** 
	 * max fdl paper width
	 */
	static protected final int MAX_FDL_PAPER_WIDTH = 21000; //2100
	/**
	 * max fdl paper height
	 */
	static protected final int MAX_FDL_PAPER_HEIGHT = 21000; //2100
	/**
	 * fdl A4 size paper width
	 */
	static protected final int MIN_FDL_PAPER_WIDTH = 2970; //297;
	/**
	 * fdl A4 size paper height
	 */
	static protected final int MIN_FDL_PAPER_HEIGHT = 2100; //210;
	
	/**
	 * use for bend distance of the data link from the control link
	 */
	static protected final int BEND_OFFSET = 20;
	/**
	 * spare border size for both width and height
	 */
	static protected final int BORDER_SIZE = 200; 	
	
	/**
	 * this is used to control the distance between nodes in the graph
	 */
	static protected final int NODE_SIZE = 300; 
	
	
	/**
	 * The graph can be moved horizontally by this value.
	 */
	static protected final int GRAPH_OFFSET_X = 150;
	
	/**
	 * The graph can be moved vertically by this value.
	 */
	static protected final int GRAPH_OFFSET_Y = 170;
	
	/**
	 * The minimal distance between adjacent layers for the layout.
	 */
	static protected final double MIN_EDGE_DISTANCE = 10;
	
	/**
	 * The minimal distance between adjacent layers for the layout.
	 */
	static protected final double MIN_LAYER_DISTANCE = 0;


	/**
	 * The graph needed by yFiles layouter.
	 */
	protected AutoLayoutGraph ivGraph = new AutoLayoutGraph();

	/**
	 * A collection of the fdl DataFlow and ControlFlow.
	 */
	protected List ivLinks = new ArrayList();
	/**
	 * A map that directly maps the fdl node to a yFiles node
	 */
	protected Map ivNodes = new HashMap();
	

	/**
	 * The LayoutManager constructor method. 
	 */
	public AbstractFdlLayoutManager() {
		super();
	}
	

	/**
	* Creates the nodes for the yFiles graph.
	* 
	* @param modelNodeList collection of fdl nodes and links 
	*/
	protected void createNodes(List modelNodeList) {
		LoggingUtil.traceEntry(this, "createNodes(List modelNodeList)");
		Iterator iter = modelNodeList.iterator();			
		while (iter.hasNext()) {
			Object graphObject = iter.next();
			if (graphObject instanceof Activity || 
					graphObject instanceof Source || 
						graphObject instanceof Sink) {
							
				//create node for graph and add fdl node and yFiles node to nodes map			

				Node node = ivGraph.createNode();
				ivGraph.setSize(node, NODE_SIZE, NODE_SIZE);
				ivGraph.setLocation(node, 0, 0);
				
				ivNodes.put(graphObject, node);
				
				//for each block, do its own layout
				if(graphObject instanceof Block){
					new FdlBlockGraphLayoutManager((Block)graphObject).doLayout();
				}

			}
			else if (graphObject instanceof ControlFlow || graphObject instanceof DataFlow) {
				//add to links collection
				ivLinks.add(graphObject);
			}
		}
		LoggingUtil.traceExit(this, "createNodes(List modelNodeList)");
	}


	
	/**
	* Creates yFiles edges from fdl DataFlow and ControlFlow.
	* 
	* @param linkList
	*/
	protected void createEdges() {	

		Iterator edgeList = ivLinks.iterator();		

		for(Iterator it = ivLinks.iterator(); it.hasNext(); ) {
			Node source = null;
			Node target = null;
			Object link = it.next();
			if(link instanceof ControlFlow){
				source = (Node)ivNodes.get(((ControlFlow)link).getFromActivity());
				target = (Node)ivNodes.get(((ControlFlow)link).getToActivity());				
			
			}else if(link instanceof DataFlow){
				Source fdlSrc = ((DataFlow)link).getSource();
				if(fdlSrc != null){
					source = (Node)ivNodes.get(fdlSrc);	
				}else{
					source = (Node)ivNodes.get(((DataFlow)link).getFromActivity());
				}
				Sink fdlSink = ((DataFlow)link).getSink();
				if(fdlSink != null){
					target =  (Node)ivNodes.get(fdlSink);
				}else{
					target = (Node) ivNodes.get(((DataFlow)link).getToActivity());				
				}				
			}
			
			if(source != null && target != null){
				ivGraph.createEdge(source, target);
			}		
		}
		
	}

	
	/**
	 * Applies the hierarchic layout algorithm to the graph.
	 * @param orientation
	 * @param routerStyle
	 * @return LayoutGraph
	 */
	protected LayoutGraph doHierarchicLayout(byte orientation, byte routerStyle) {
		LoggingUtil.traceEntry(this, "doHierarchicLayout(byte orientation, byte routerStyle)");	
			
		HierarchicLayouter layouter = new HierarchicLayouter();
		layouter.setRoutingStyle(routerStyle);
	
		OrientationLayouter oLayouter = (OrientationLayouter)layouter.getOrientationLayouter();
		oLayouter.setOrientation(orientation);
		
		layouter.setMinimalLayerDistance(MIN_LAYER_DISTANCE);

		layouter.setMinimalEdgeDistance(MIN_EDGE_DISTANCE);
		
		layouter.setSubgraphLayouterEnabled(false);
		
						
		new BufferedLayouter(layouter).doLayout(ivGraph);
		
		LoggingUtil.traceExit(this, "doHierarchicLayout(byte orientation, byte routerStyle)");
		return ivGraph;
	}
	
	
	public void doLayout(){
		LoggingUtil.traceEntry(this, "doLayout()");
		
		//use this specific layout style only
		doLayout(CefLiterals.ACTION_ID_ORTHOGONAL_FLOW_LAYOUT_LEFT_RIGHT);
		
		LoggingUtil.traceExit(this, "doLayout()");
	}
	

	/**
	 * Applies the specified layout to the model.
	 * @param layoutStyle
	 * @param diagram 
	 */
	protected void doLayout(String layoutStyle) {		
		LoggingUtil.traceEntry(this, "doLayout(String layoutStyle)");
		ivGraph.clear();
		
		generateGraphFromModel();
			
		basicDoLayout(layoutStyle);
		
		//update result
		updateResults();
		
		LoggingUtil.traceExit(this, "doLayout(String layoutStyle)");
			
	}
	
	/**
	 * Applies the specified layout for the graph.
	 * @param layoutStyle
	 */
	protected void basicDoLayout(String layoutStyle){
		LoggingUtil.traceEntry(this, "basicDoLayout(String layoutStyle)");
		
		if (CefLiterals.ACTION_ID_ORTHOGONAL_FLOW_LAYOUT_TOP_BOTTOM.equals(layoutStyle)) {
			doHierarchicLayout(OrientationLayouter.TOP_TO_BOTTOM, HierarchicLayouter.ROUTE_ORTHOGONAL);
		}

		else if (CefLiterals.ACTION_ID_ORTHOGONAL_FLOW_LAYOUT_LEFT_RIGHT.equals(layoutStyle)) {
			doHierarchicLayout(OrientationLayouter.LEFT_TO_RIGHT, HierarchicLayouter.ROUTE_ORTHOGONAL);
		}

		else if (CefLiterals.ACTION_ID_ORTHOGONAL_FLOW_LAYOUT_BOTTOM_TOP.equals(layoutStyle)) {
			doHierarchicLayout(OrientationLayouter.BOTTOM_TO_TOP, HierarchicLayouter.ROUTE_ORTHOGONAL);
		}
		else if (CefLiterals.ACTION_ID_ORTHOGONAL_FLOW_LAYOUT_RIGHT_LEFT.equals(layoutStyle)) {
			doHierarchicLayout(OrientationLayouter.RIGHT_TO_LEFT, HierarchicLayouter.ROUTE_ORTHOGONAL);
		}
		else if (CefLiterals.ACTION_ID_POLYLINE_FLOW_LAYOUT_TOP_BOTTOM.equals(layoutStyle)) {
			doHierarchicLayout(OrientationLayouter.TOP_TO_BOTTOM, HierarchicLayouter.ROUTE_POLYLINE);
		}
		else if (CefLiterals.ACTION_ID_POLYLINE_FLOW_LAYOUT_LEFT_RIGHT.equals(layoutStyle)) {
			doHierarchicLayout(OrientationLayouter.LEFT_TO_RIGHT, HierarchicLayouter.ROUTE_POLYLINE);
		}
		else if (CefLiterals.ACTION_ID_POLYLINE_FLOW_LAYOUT_RIGHT_LEFT.equals(layoutStyle)) {
			doHierarchicLayout(OrientationLayouter.RIGHT_TO_LEFT, HierarchicLayouter.ROUTE_POLYLINE);
		}
		else if (CefLiterals.ACTION_ID_POLYLINE_FLOW_LAYOUT_BOTTOM_TOP.equals(layoutStyle)) {
			doHierarchicLayout(OrientationLayouter.BOTTOM_TO_TOP, HierarchicLayouter.ROUTE_POLYLINE);;
		}

		ivGraph.moveGraph(GRAPH_OFFSET_X, GRAPH_OFFSET_Y);
		
		LoggingUtil.traceExit(this, "basicDoLayout(String layoutStyle)");
	}
		

		
	/**
	* Creates graph for the fdl model
	* 
	*/
	protected void generateGraphFromModel() {		
		LoggingUtil.traceEntry(this, "generateGraphFromModel()");
		
		createNodes(getModelNodes());
		createEdges();	
		
		LoggingUtil.traceExit(this, "generateGraphFromModel()");		
	}
	
	/**
	 * Subclass must implement this method to return the model nodes in the graph to be layed outh.
	 * @return
	 */
	protected abstract List getModelNodes();
	
		
	/**
	 * Updates links layouts.
	 */
	protected void updateLinkResults()
	{
		LoggingUtil.traceEntry(this, "updateLinkResults()");
		//here we only need to set bend point for data flow. 
		//caculate Dataflow bend point by using the 1/2 of source and 1/2 of target position offset by BEND_OFFSET.
		for(Iterator it = ivLinks.iterator(); it.hasNext(); ){
			Object link = it.next();
			if(link instanceof DataFlow){
				int bendX = 0;
				int bendY = 0;
				int srcX = 0, srcY = 0, tarX = 0, tarY = 0;
				Activity defaultActivity = ((DataFlow)link).getDefaultActivity(); //this is case link goes to itself
				Activity loopActivity = ((DataFlow)link).getLoop(); //this is case link is loop to itself 
				if(defaultActivity != null){
					bendX = defaultActivity.getScreenPosition().getXpos().intValue();
					bendY = defaultActivity.getScreenPosition().getYpos().intValue() + 100;
				}else if(loopActivity != null){
					bendX = loopActivity.getScreenPosition().getXpos().intValue();
					bendY = loopActivity.getScreenPosition().getYpos().intValue() + 100;					
				}else{
					Source src = ((DataFlow)link).getSource();
					if(src != null){				
						srcX = src.getScreenPosition().getXpos().intValue();
						srcY = src.getScreenPosition().getYpos().intValue();
					}else{
						Activity activity = ((DataFlow)link).getFromActivity();
						if(activity != null){
							srcX = activity.getScreenPosition().getXpos().intValue(); 
							srcY = activity.getScreenPosition().getYpos().intValue();
						}
					}
					
					Sink sink =  ((DataFlow)link).getSink();
					if(sink != null){				
						tarX = sink.getScreenPosition().getXpos().intValue();
						tarY = sink.getScreenPosition().getYpos().intValue();
					}else{
						Activity activity = ((DataFlow)link).getToActivity();
						if(activity != null){
							tarX = activity.getScreenPosition().getXpos().intValue(); 
							tarY = activity.getScreenPosition().getYpos().intValue();
						}
					}
					if(Math.abs(srcX - tarX) < BEND_OFFSET){
						bendX = srcX + BEND_OFFSET;
						bendY = (srcY + tarY) / 2;
					}else if(Math.abs(srcY - tarY) < BEND_OFFSET){
						bendX = (srcX + tarX) / 2;
						bendY = srcY + BEND_OFFSET;
					}else{
						bendX = (srcX + tarX) / 2;
						bendY = (srcY + tarY) / 2 + BEND_OFFSET;
					}
				}
				
				ScreenPosition bendPoint = ModelFactory.eINSTANCE.createScreenPosition();
				bendPoint.setXpos(new Integer(bendX));
				bendPoint.setYpos(new Integer(bendY));				
				((DataFlow)link).getBendPoints().add(bendPoint);
			}
		}

		LoggingUtil.traceExit(this, "updateLinkResults()");
	}

	/**
	* Updates the fdl nodes' position with node layouts.
	*/
	protected void updateNodeResults()	{
		
		LoggingUtil.traceEntry(this, "updateNodeResults()");
				
		int graphWidth = getMaxWidthAndMaxHeight(ivNodes)[0] + BORDER_SIZE;
		int graphHeight = getMaxWidthAndMaxHeight(ivNodes)[1] + BORDER_SIZE;
		int paperWidth = graphWidth;
		int paperHeight = graphHeight;

		
		float xscale = 1, yscale = 1;	


		if(graphWidth < MIN_FDL_PAPER_WIDTH){
			graphWidth = MIN_FDL_PAPER_WIDTH;
			paperWidth = graphWidth;
		}else if(graphWidth > MAX_FDL_PAPER_WIDTH){
			xscale = ((float)MAX_FDL_PAPER_WIDTH) / graphWidth;
			//System.out.println("xscale = " + xscale);
			paperWidth = MAX_FDL_PAPER_WIDTH;			
		}		

		if(graphHeight < MIN_FDL_PAPER_HEIGHT){
			graphHeight = MIN_FDL_PAPER_HEIGHT;
			paperHeight = graphHeight;
		}else if(graphHeight > MAX_FDL_PAPER_HEIGHT){
			yscale = ((float)MAX_FDL_PAPER_HEIGHT) / graphHeight;
			//System.out.println("yscale = " + yscale);
			paperHeight = MAX_FDL_PAPER_HEIGHT;
		}
		
		PaperSize paperSize = ModelFactory.eINSTANCE.createPaperSize();
		paperSize.setWidth(new Integer(paperWidth));
		paperSize.setHeight(new Integer(paperHeight));
		setPaperSize(paperSize);
		
		int halfGraphWidth = graphWidth / 2;
		int halfGraphHeight = graphHeight / 2;
		//System.out.println("halfGraphWidth = " + halfGraphWidth);
		//System.out.println("halfGraphHeight = " + halfGraphHeight);
		
		int x = 0, y = 0;
		for (Iterator it = ivNodes.keySet().iterator(); it.hasNext(); ) {
			Object fdlNodeModel = it.next();
			Object graphObject = ivNodes.get(fdlNodeModel);
			NodeLayout nodeLayout = ivGraph.getNodeLayout((Node) graphObject);
			x = (int) nodeLayout.getX();
			y = (int) nodeLayout.getY();
			ScreenPosition layout = ModelFactory.eINSTANCE.createScreenPosition();
			layout.setXpos(new Integer(convertX(x, halfGraphWidth, xscale)));
			layout.setYpos(new Integer(convertY(y, halfGraphHeight, yscale)));			
			if(fdlNodeModel instanceof Activity){
				((Activity)fdlNodeModel).setScreenPosition(layout);
			}else if(fdlNodeModel instanceof Source){
				((Source)fdlNodeModel).setScreenPosition(layout);
			}else if(fdlNodeModel instanceof Sink){
				((Sink)fdlNodeModel).setScreenPosition(layout);
			}
		}
		
		LoggingUtil.traceExit(this, "updateNodeResults()");		
	}
	
	/**
	 * Subclass to call to set paper size 
	 * @param paperSize
	 */
	
	protected abstract void setPaperSize(PaperSize paperSize);
	
	/**
	 * Returns the max x and y coordinates for the list of nodes' graph object
	 * @param nodes
	 * @return
	 */
	private int[] getMaxWidthAndMaxHeight(Map nodes){
		LoggingUtil.traceEntry(this, "getMaxWidthAndMaxHeight(Map nodes");
		int maxX = 0,  maxY = 0;
		for(Iterator it =  nodes.keySet().iterator(); it.hasNext(); ) {
			Object fdlNode = it.next();
			Object graphObject = nodes.get(fdlNode);
			NodeLayout nodeLayout = ivGraph.getNodeLayout((Node) graphObject);
			int x = (int) nodeLayout.getX();
			int y = (int) nodeLayout.getY();	
			if(x > maxX){
				maxX= x;
			}
			if(y > maxY){
				maxY = y;
			}
		}
		LoggingUtil.traceExit(this, "getMaxWidthAndMaxHeight(Map nodes", 
			"maxX --> "  + maxX + " maxY --> " + maxY);
		return new int[] {maxX, maxY};
	}
	
	/**
	 * Coverts x coordinate of yFile layout to FDL graph x coordinate. 
	 * Apply the scale factor to x value.
	 * @param x
	 * @param maxFdlWidth
	 * @param scale
	 * @return
	 */
	private int convertX(int x, int halfGraphWidth, float scale){
		LoggingUtil.traceEntry(this, "convertX(int x, int halfGraphWidth, float scale)", 
			"x --> " + x + " halfGraphWidth --> " + halfGraphWidth + " scale --> " + scale);
			
		// X = -a + x where a = graphWidth/2
		//System.out.println("x = " + x);
		int X = -halfGraphWidth + x;
		//System.out.println("X = " + (int) (X * scale));
		LoggingUtil.traceExit(this, "convertX(int x, int halfGraphWidth, float scale)",
			"result --> " + (int) (X * scale));
		return  (int) (X * scale);
	}
	
	/**
	 * Converts y coordinate of yFile layout to FDL graph y coordinate.
	 * Apply the scale factor to y value.
	 * @param y
	 * @param graphHeight
	 * @param scale
	 * @return
	 */
	private int convertY(int y, int halfGraphHeight, float scale){
		LoggingUtil.traceEntry(this, "convertY(int y, int halfGraphHeight, float scale)", 
			"y --> " + y + " halfGraphHeight --> " + halfGraphHeight + " scale --> " + scale);
		// Y = b - y whre b = graphHeight / 2
		//System.out.println("y = " + y);
		int Y = halfGraphHeight - y;
		//System.out.println("Y = " + (int) (Y * scale));
		LoggingUtil.traceExit(this, "convertY(int y, int halfGraphHeight, float scale)",
			"result --> " + (int) (Y * scale));
		return (int)(Y * scale);
	}
	/**
	 * Updates the layout results to the fdl node's x and y coordinates.
	 */
	protected void updateResults() {
		updateNodeResults();
		updateLinkResults();
	}

}


