package com.ibm.btools.te.ilm.heuristics.fdl.util;
/*
* Licensed Materials - Property of IBM
* 5724-I74, 5724-I75
*  Copyright IBM Corporation 2003, 2009. All rights reserved.
* US Government Users Restricted Rights - Use, duplication or disclosure
* restricted by GSA ADP Schedule Contract with IBM Corp.
*/
import java.util.List;

import com.ibm.btools.bom.model.artifacts.Property;
import com.ibm.btools.bom.model.artifacts.Type;
import com.ibm.btools.bom.model.processes.activities.Action;
import com.ibm.btools.bom.model.processes.activities.Activity;
import com.ibm.btools.bom.model.processes.activities.InputPinSet;
import com.ibm.btools.bom.model.processes.activities.ObjectPin;
import com.ibm.btools.bom.model.processes.activities.PinSet;
import com.ibm.btools.fdl.model.DataStructure;

/**
 * A name registry interface.
 * This name registry provides funtions to produce valid names and register names for different types of FDL constructs.
 */
public interface FdlNameRegistry {
	
	public static final String copyright = "Licensed Material - Property of IBM " + " 5724-I74, 5724-I75 " + "(C) Copyright IBM Corporation 2003, 2009. All Rights Reserved. " + "U.S. Government Users Restricted Rights - Use, duplication or disclosure " + "restricted by GSA ADP Schedule Contract with IBM Corp.".intern(); //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$ //$NON-NLS-4$ //$NON-NLS-5$
	/**
	 * Returns activity name for the given aciton.
	 * It searches the name registry and if found a name already registered for the action, 
	 * it returns that name, otherwise generates one and stores it in the registry. 
	 * This method will use the action's full name and strip any reserved characters and words, 
	 * and make the name unique within the current Process transformation. 
	 * 
	 * @param action	The action that the activity name is to be retrieved for or generated from. 
	 * @return String - The activity name retrieved or generated
	 */
	public abstract String getActivityName(Action action);
	
	/**
	 * Returns activity name for the given two parts.
	 * It searched the name registry and if found a name already registered for the mainPart, 
	 * it returns that name and append the suffix, otherwise generate one and store it in the name registry. 
	 * This method will strip any reserved character and words from the main part 
	 * and shorten the length if necessary. It guarantees the name is unique within the current Process transformation.
	 *  
	 * @param mainPart	The activity name's main part. 
	 * @param suffix	The suffix that need to be appended to the name. 
	 * 					The suffix need to be the same length between differnt methods calls for the same main part.
	 * @return String - The activity name retrieved or generated
	 */
	public abstract String getActivityName(String mainPart, String suffix);
	
	/**
	 * Returns activity name for the given pinSet.
	 * It searches the name registry and if found a name already registered for the pinSet's action, 
	 * it returns that name and append the index number of the pinSet if the action has more than one pinSets, 
	 * otherwise generate a new one and store it in the name registry. 
	 * This method will use the action's full name, and strip any reserved character and words  
	 * and shorten the length if necessary. It guarantees the name is unique within the current Process transformation.
	 *  
	 * @param pinSet	The pin set that the name will be retrieved for or generated from.  
	 * @return String - The activity name retrieved or generated
	 */
	public abstract String getActivityName(PinSet pinSet);
	
	/**
	 * Returns program name for the given two parts.
	 * It searches the name registry first and if found a name registered for the mainPart, 
	 * it returns that name and append the suffix. 
	 * This method will strip any reserved character for the main part 
	 * and shorten the mainPart length if necessary. 
	 * It guarantees the name is unique within one transformation session.
	 *  
	 * @param mainPart	The program name's main part. 
	 * @param suffix	The suffix that need to be appended to the name. 
	 * 					The suffix length should be the same between different calls to this method.
	 * @return String
	 */
	public abstract String getProgramName(String mainPart, String suffix);
	
	/**
	 * Returns program name for the given InputPinSet.
	 * It searches the name registry first and if found a name already registered for the inputPinSet's action,
	 * it returns that name and append the index of the inputPinSet. 
	 * This method will use the action's full name and strip any reserved character and words, 
	 * and make the name unique within one transformation session.
	 *  
	 * @param inputPinSet	The InputPinSet that the name is to be retrieved for or generated from. 
	 * @return String
	 */
	public abstract String getProgramName(InputPinSet inputPinSet);
	
	/**
	 * Returns process name for the given activity and index.
	 * It searches the name registry first and if found a name already registered for the activity,
	 * it returns that name and append the index. 
	 * This method will use the activity's full name and strip any reserved character, 
	 * and make the name unique within one transformation session. 
	 * 
	 * @param activity	The activity that the name will be retrieved for or generated from. 
	 * @param index		The index that need to be appended to the name or null.
	 * @return String
	 */
	public abstract String getProcessName(Activity activity, Integer index);
	
	/**
	 * Returns process category name for the given activity.
	 * It searches the name registry first and if found a process categoty name already registered for the activity,
	 * it returns that name, otherwise, it generates a new one. 
	 * This method will use the activity's full package name and strip any reserved character, and shorten the length is necessary, 
	 * and make the name unique within one transformation session. 
	 * @param activity  The activity the process category name can be retrieved for or generated from. 
	 * @return String
	 */
	public abstract String getProcessCategoryName(Activity activity);
	/**
	 * Returns data structure name for the given pinSet.
	 * It searches the name registry first and if found a data structure name already registered for the pinSet,
	 * it returns that name, otherwise, it generates a new one. 
	 * This method will use the pinSet's full name and strip any reserved character, and shorten the length is necessary, 
	 * and make the name unique within one transformation session.
	 *  
	 * @param pinSet  The pinSet the data structure is retrieved for or generated from. 
	 * @return String
	 */
	public abstract String getDataStructureName(PinSet pinSet);
	/**
	 * Returns data structure name for the given type.
	 * It searches the name registry first and if found a data structure name already registered for the type,
	 * it returns that name, otherwise, it generates a new one. 
	 * This method will use the type's full name and strip any reserved character, shorten the length is necessary, 
	 * and make the name unique within one transformation session.
	 *  
	 * @param type  The type the data structure is retrieved for or generated from. 
	 * @return String
	 */
	public abstract String getDataStructureName(Type type);
	
	/**
	 * Returns a key that is unique for the composition of pin types and multiplicity of the pins in the pinSet. 
	 * This key can be used to classify identical structure composition of different pinSets.
	 * 
	 * @param pinSet	The PinSet that the key is to be computed from. 
	 * @return String 
	 */
	public abstract String computeDataStructureKey(PinSet pinSet);
	
	/**
	 * Returns s trancated to fit the length of FDL description restriction if necessary. 
	 */
	public abstract String getDescription(String s);
	
	/**
	 *  Returns s trancated to fit the length of FDL documentation restriction if necessary.  
	 */
	public abstract String getDocumentation(String s);
	
	/**
	 *  Returns a string extracted and concatenated from the list. 
	 *  The whole string may be trancated to fit the length of FDL documentation restriction if necessary.
	 * 
	 * @param list a list of description strings
	 * @return String  
	 */
	public abstract String getDocumentation(List list);
	
	/**
	 * Cleans up registry for transformation of one process.
	 * The activity name registry will be cleared.
	 */
	public abstract void resetForProcess();
	
	/**
	 * Cleans up registry for one transformation session.
	 * All name registry will be cleared.
	 */
	public abstract void resetForSession();
	
	/**
	 *  Returns the data structure memeber name for the objPin.
	 *  This method first try to search if a name is already registered for th objPin within the ds, 
	 *  if not found, it uses the objPin type's name and strip reserved characters and make it unique within the ds.
	 * 
	 * @param ds		the data structure that the objPin is part of.
	 * @param objPin 	the object pin that the member name is retrieved for or generated from.
	 * @return String
	 */
	public abstract String getDataStructureMemberName(
		DataStructure ds,
		ObjectPin objPin);
	
	/**
	 *  Returns the data structure memeber name for the property.
	 *  This method first try to search if a name is already registered for th property within the datastructure that created out of the type, 
	 *  if not found, it uses the property's name, strip reserved characters and make it unique within the datastructure.
	 *  this method can be used when the property is from a super type and need to to be flatten into <code>type</code>.
	 * 
	 * @param property	the property that the name is retrieved for or generated from. 
	 * @param type		the type that the datastructure is generated from in which the property is the member.
	 * @return String 
	 */
	public abstract String getDataStructureMemberNameForType(
		Property property,
		Type type);
	
	/**
	 *  Returns the data structure memeber name for the property.
	 *  This method first try to search if a name is already registered for th property 
	 *  within the datastructure that created out of the type that containing the property, 
	 *  if not found, it uses the property's name, strip reserved characters and make it unique within the datastructure.
	 * 
	 * @param property	the property that the member name is retrieved for or generated from.
	 * @return String
	 */
	public abstract String getDataStructureMemberName(Property property);
	
	/**
	 * Creates a unqiue Data Structure name based on the structural content of the data structure
	 * @param dsKey - key created from computeDataStructureKey
	 * @return String
	 */
	public abstract String getUniqueDataStructureName(String dsKey);

	/**
	 * Creates a unqiue Program name based on the input and output dataStructure
	 * @param pKey - combination of the name of the input and output dataStructure
	 * @return String
	 */
	public abstract String getUniqueProgramName(String pKey);

}