package com.ibm.btools.te.ilm.heuristics.fdl.util;
/*
* Licensed Materials - Property of IBM
* 5724-I74, 5724-I75
*  Copyright IBM Corporation 2003, 2009. All rights reserved.
* US Government Users Restricted Rights - Use, duplication or disclosure
* restricted by GSA ADP Schedule Contract with IBM Corp.
*/

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.StringTokenizer;

import org.eclipse.emf.ecore.EObject;

import com.ibm.btools.bom.model.artifacts.LiteralInteger;
import com.ibm.btools.bom.model.artifacts.LiteralUnlimitedNatural;
import com.ibm.btools.bom.model.artifacts.NamedElement;
import com.ibm.btools.bom.model.artifacts.Package;
import com.ibm.btools.bom.model.artifacts.PackageableElement;
import com.ibm.btools.bom.model.artifacts.Property;
import com.ibm.btools.bom.model.artifacts.Type;
import com.ibm.btools.bom.model.artifacts.ValueSpecification;
import com.ibm.btools.bom.model.models.ProcessModel;
import com.ibm.btools.bom.model.processes.activities.Action;
import com.ibm.btools.bom.model.processes.activities.Activity;
import com.ibm.btools.bom.model.processes.activities.InputPinSet;
import com.ibm.btools.bom.model.processes.activities.ObjectPin;
import com.ibm.btools.bom.model.processes.activities.OutputPinSet;
import com.ibm.btools.bom.model.processes.activities.PinSet;
import com.ibm.btools.fdl.model.DataStructure;
import com.ibm.btools.te.ilm.heuristics.helper.BomUtils;
import com.ibm.btools.te.ilm.heuristics.helper.ProcessUtil;

/**
 * A FdlNameRegistry implementation to produce valid names and register names for different types of FDL constructs.
 */
public class FdlNameRegistryImpl implements  FdlNameRegistry {
	
	public static final String copyright = "Licensed Material - Property of IBM " + " 5724-I74, 5724-I75 " + "(C) Copyright IBM Corporation 2003, 2009. All Rights Reserved. " + "U.S. Government Users Restricted Rights - Use, duplication or disclosure " + "restricted by GSA ADP Schedule Contract with IBM Corp.".intern(); //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$ //$NON-NLS-4$ //$NON-NLS-5$
	
	/*
	 * predefined maximum length, reserved characters and words
	 */

	static final int MAX_OBJECT_NAME_LENGTH = 32;
	//static final int MAX_PROCESS_NAME_LENGTH = 63;
	static final int MAX_PROCESS_NAME_LENGTH = 32;
	static final int MAX_SYMBOL_NAME_LENGTH = 32;
	//static final int MAX_DESCRIPTION_LENGTH = 250;
	static final int MAX_DESCRIPTION_LENGTH = 254;
	static final int MAX_DOCUMENTATION_LENGTH = 4096;

	static final int MAX_ACTIVITY_NAME_LENGTH = MAX_SYMBOL_NAME_LENGTH;
	static final int MAX_PROGRAM_NAME_LENGTH = MAX_OBJECT_NAME_LENGTH;
	static final int MAX_PROCESS_CATEGORY_NAME_LENGTH = MAX_OBJECT_NAME_LENGTH;
	static final int MAX_DATA_STRUCTURE_NAME_LENGTH = MAX_OBJECT_NAME_LENGTH;
	static final int MAX_DATA_STRUCTURE_MEMBER_NAME_LENGTH = MAX_SYMBOL_NAME_LENGTH;

	static final char[] SYMBOL_NAME_CHAR_NOT =
				{'!', '\'','[', ']', '*', '+' , ',', '-', '.', ';',
				 '/', ':', '<', '=', '>', '(', ')', '\\', '^', '"', ' ', '\'' };			 
			 
	static final char[] PROCESS_NAME_CHAR_NOT =
				{'*', '?', '"', ';', ':', '.', '$', '\''};

	/* characters are not allowed in MQ Workflow Buildtime, but not documented */			
	static final char[] DATA_STRUCTURE_NAME_CHAR_NOT =
					{'*', '?', '"', ';', ':', '.', '\''};	
	static final char[] DESCRIPTION_CHAR_NOT = {'"', '\'', '\t'};		
	static final char[] DOCUMENTATION_CHAR_NOT = {'"', '\'', '\t'};
						
	static final String[] SYMBOL_NAME_WORD_NOT =
				{"AND", "IS", "LOWER", "MOD", "NOT", "NULL", "OR", 
				"SUBSTR", "UPPER", "VALUE", "_BLOCK", "_STRUCT"};

	static final String DEFAULT_DATA_STRUCTURE_NAME_PREFIX="DataStructure";
	static final String DEFAULT_PROGRAM_NAME_PREFIX="Program";
	
	private  Map activityNameRegistry = new HashMap();
	private  Map programNameRegistry = new HashMap();	
	private  Map processNameRegistry = new HashMap();	
	private  Map processCategoryNameRegistry = new HashMap();	
	private  Map dataStructureNameRegistry = new HashMap();
	private  Map dataStructureMemberNameRegistry = new HashMap();	
	
	
	/**
	 * Strips any character that are not allowed as symbol name.
	 * 
	 * @param name	the name to be examined.
	 * @return String
	 */
	private  String restrictSymbolName(String name){
			name = name.trim();
			name = stripChars(name, SYMBOL_NAME_CHAR_NOT);
			//name cannot start from digit	
			for(int i = 0, size= name.length(); i< size; i++ ){
				if(!Character.isDigit(name.charAt(i))){
					name = name.substring(i);
					break;
				}
			}
			return name;
		}
	
	/**
	 * Strips any character that are not allowed as data structure name.
	 * 
	 * @param name	the name to be examined.
	 * @return String
	 */
	private String  restrictDataStructureMemberName(String name){
				name = name.trim();
				name = stripChars(name, SYMBOL_NAME_CHAR_NOT);
				//name cannot start from underscore	
				// Ray:Cannot have leading digit
				for(int i = 0, size= name.length(); i< size; i++ ){
					if(!(name.charAt(i) == '_') && (!Character.isDigit(name.charAt(i)))){
						name = name.substring(i);
						break;

					} 
				}
				return name;
			}	

	/**
	 * Returns activity name for the given action.
	 * It searches the name registry and if found a name already registered for the action, 
	 * it returns that name, otherwise generate one and store it in the registry. 
	 * This method will use the action's full name and strip any reserved characters and words, 
	 * and make the name unique within the current Process transformation. 
	 * 
	 * @param action	the action that the activity name is to be retrieved for or generated from. 
	 * @return String - the activity name retrieved or generated
	 */
	public  String getActivityName(Action action){
		return getActivityName(action, null, 0);
	
	}

/*
	private String getActivityNameInSubContext(String subContextKey, String nameKey) {
		Map subContext = getActivityNameSubContext(subContextKey);
		return (String)subContext.get(nameKey);
	}
*/
	private Map getActivityNameSubContext(String subContextKey) {
		Map subContext = (Map)activityNameRegistry.get(subContextKey);
		if (subContext == null) {
			subContext = new HashMap();
			activityNameRegistry.put(subContextKey,subContext);
		}
		return subContext;
	}

	/**
	 * Returns activity name for the given action.
	 * It searches the name registry and if found a name already registered for the action, 
	 * it returns that name and append the index as suffix, otherwise generate one and store it in the registry. 
	 * This method will use the action's full name and strip any reserved characters and words, 
	 * and make the name unique within the current Process transformation. 
	 * 
	 * @param action	the action that the activity name is to be retrieved for or generated from. 
	 * @param index		the index that need to be appended to the name or null. 
	 * @param maxIndexLength	the maximum length of the index might be
	 * @return String - the activity name retrieved or generated
	 */
	private String getActivityName(Action action, String indexString, int maxIndexLength ){
		String name;
		String nameKey = getFullName(action);
		String containerKey = getFullName(action.getInStructuredNode());
		Map subContext = getActivityNameSubContext(containerKey);
		int maxLength;
		maxLength = MAX_ACTIVITY_NAME_LENGTH - maxIndexLength;	
		if (indexString!=null){
			maxLength-=1+indexString.length();
		}				
		//if((name =(String) activityNameRegistry.get(nameKey)) != null){
		if((name =(String) subContext.get(nameKey)) != null){
			if (name.length()>maxLength){
				name = name.substring(0,maxLength);	
				name=name.trim();							
			}			
			name = (indexString == null)? name: name + "_" + indexString;	
			return name;
		}
		
		//AR Use the CBA name instead
//		if (action instanceof CallBehaviorAction) {
//			name = ((CallBehaviorAction)action).getBehavior().getName().trim();
//		} else {
//			name = action.getName().trim();
//		}
		name = action.getName().trim(); 
		name = stripWords(name, SYMBOL_NAME_WORD_NOT); 
		name = restrictSymbolName(name);

		if(name.length()> maxLength){
		   //trucate to make it in the range
		   name = name.substring(0,maxLength);	
		   name=name.trim();				
		}	
	 
		name = getUniqueName(name, 
					//activityNameRegistry.values(),
					subContext.values(), 
					maxLength);
		
		//activityNameRegistry.put(nameKey, name);
		subContext.put(nameKey,name);

		name = (indexString == null)? name: name + "_" + indexString;

		return name;
	}



	/**
	 * Returns activity name for the given two parts.
	 * It searched the name registry and if found a name already registered for the mainPart, 
 	 * it returns that name and append the suffix, otherwise generate one and store it in the name registry. 
	 * This method will strip any reserved character and words from the main part 
	 * and shorten the length if necessary. It guarantees the name is unique within the current Process transformation.
	 *  
	 * @param mainPart  the activity name's main part. 
	 * @param suffix    the suffix that need to be appended to the name. 
	 * 					the suffix need to be the same length between differnt methods calls.
	 * @return
	 */	
	public  String getActivityName(String mainPart, String suffix){
		String name;
		String nameKey = mainPart;
		if (activityNameRegistry.get(nameKey) instanceof String) {
			if((name =(String) activityNameRegistry.get(nameKey)) != null){
				name = (suffix == null)? name: name + suffix;	
				return name;
			}
		}

		int maxLength;
		maxLength = MAX_ACTIVITY_NAME_LENGTH - suffix.length();	

		name = mainPart.trim(); 
		name = stripWords(name, SYMBOL_NAME_WORD_NOT); 
		name = restrictSymbolName(name);
		if(name.length()> maxLength){
		   //trucate to make it in the range
		   name = name.substring(0,maxLength);	
		   name=name.trim();				
		}	
	 
	 	// Temporary fix ... flatten the registry values
	 	Collection values = activityNameRegistry.values();
	 	ArrayList flattenVals = new ArrayList();
	 	for (Iterator it=values.iterator(); it.hasNext();){
	 		Object val = it.next();
	 		if (val instanceof Map) {
	 			flattenVals.addAll(((Map)val).values());
	 		} else {
	 			flattenVals.add(val);
	 		}
	 	}
		name = getUniqueName(name, 
					//activityNameRegistry.values(),
					flattenVals, 
					maxLength);
		
		activityNameRegistry.put(nameKey, name);

		name = (suffix == null)? name: name + suffix;

		return name;
		}


	/**
	 * Returns activity name for the given pinSet.
	 * It searches the name registry and if found a name already registered for the pinSet's action, 
	 * it returns that name and append the index number of the pinSet, 
	 * otherwise generate a new one and store it in the name registry. 
	 * This method will use the action's full name, and strip any reserved character and words  
	 * and shorten the length if necessary. It guarantees the name is unique within the current Process transformation.
	 *  
	 * @param pinSet	the pin set that the name will be retrieved for or generated from.  
	 * @return String - the activity name retrieved or generated
	 */
	public  String getActivityName(PinSet pinSet){
		Action action;
		int index;
		int numberOfPinSet; 
		if(pinSet instanceof InputPinSet){
			action = ((InputPinSet)pinSet).getAction();			
			index = action.getInputPinSet().indexOf(pinSet) + 1;
			numberOfPinSet = action.getInputPinSet().size();	
		}
		else{
			action = ((OutputPinSet)pinSet).getAction();
			index = action.getOutputPinSet().indexOf(pinSet) + 1;
			numberOfPinSet = action.getOutputPinSet().size();	
		}
		
		String name = null;
		if(numberOfPinSet > 1){
			name = getActivityName(action, String.valueOf(index), String.valueOf(numberOfPinSet).length());		
		}else{
			name = getActivityName(action, null, 0 );
		}
		return name;
	}


	/**
	 * Returns program name for the given two parts.
	 * It searches the name registry first and if found a name registered for the mainPart, 
	 * it returns that name and append the suffix. 
	 * This method will strip any reserved character for the main part 
	 * and shorten the mainPart length if necessary. 
	 * It guarantees the name is unique within one transformation session. 
	 * @param mainPart the program name's main part. 
	 * @param suffix   the suffix that need to be appended to the name. 
	 * 				   The suffix length should be the same between different calls to this method.
	 * @return String
	 */	
		public  String getProgramName(String mainPart, String suffix){
			String name;
			String nameKey = mainPart;
			if((name =(String) programNameRegistry.get(nameKey)) != null){
				name = (suffix == null)? name: name + suffix;	
				return name;
			}
	
			int maxLength;
			maxLength = MAX_PROGRAM_NAME_LENGTH - suffix.length();	
	
			name = mainPart.trim(); 
			name = stripChars(name, PROCESS_NAME_CHAR_NOT);
			if(name.length()> maxLength){
			   //trucate to make it in the range
			   name = name.substring(0,maxLength);		
			   name=name.trim();			
			}	
		 
			name = getUniqueName(name, 
						programNameRegistry.values(), 
						maxLength);
			
			programNameRegistry.put(nameKey, name);
	
			name = (suffix == null)? name: name + suffix;

			return name;
			}
			
			

	/**
	 * Returns program name for the given InputPinSet.
	 * It searches the name registry first and if found a name already registered for the inputPinSet's action,
	 * it returns that name and append the index of the inputPinSet. 
	 * This method will use the action's full name and strip any reserved character and words, 
	 * and make the name unique within one transformation session.
	 *  
	 * @param inputPinSet	the InputPinSet that the name is to be retrieved for or generated from. 
	 * @return String
	 */
	public  String getProgramName(InputPinSet inputPinSet){
		Action action = inputPinSet.getAction();
		Integer index = null;
		if(action.getInputPinSet().size() > 1){
			index = new Integer( action.getInputPinSet().indexOf(inputPinSet) + 1);		
		}
		return getProgramName(action, index);
	}

	/**
	 * Returns program name for the given action and index.
	 * It searches the name registry first and if found a name already registered for the action,
	 * it returns that name and append the index. 
	 * This method will use the action's full name and strip any reserved character and words, 
	 * and make the name unique within one transformation session.
	 *  
	 * @param action	the action that the name will be retrieved for or generated from. 
	 * @param index		the index that need to be appended to the name or null.
	 * @return String
	 */
	private  String getProgramName(Action action, Integer index){
		
			String name;
			String nameKey = getFullName(action);
			int numberOfSet = action.getInputPinSet().size();
			int maxLength = (numberOfSet >1)?	
			 (MAX_PROGRAM_NAME_LENGTH - String.valueOf(numberOfSet).length()) :
			  MAX_PROGRAM_NAME_LENGTH;		
			  
			if (index!=null){
				maxLength-=1+index.toString().length();
			}		  
			  
			  	
			if((name =(String) programNameRegistry.get(nameKey)) != null){
				if(name.length()> maxLength){
				   //trucate to make it in the range
				   name = name.substring(0,maxLength);	
				   name=name.trim();				
				}					
				name = (index == null)? name: name + "_" + index.toString();	
				return name;
			}	

		//	String name = getCanonicalName(action.getInStructuredNode(), false);
		
		//	name = "Program"; 
			name = nameKey;
				
		//	if(name.length() > maxLength){
		//		name = name.substring(0, maxLength);
		//	}
	
			name = getUniqueName(name, 
						programNameRegistry.values(), 
						maxLength);
		
			programNameRegistry.put(nameKey, name);
			
			name = (index == null)? name: name + "_" + index.toString();
			return name;
		
			}
		
		
	/**
	 * Returns process name for the given activity and index. 
	 * It searches the name registry first and if found a name already registered for the activity,
	 * it returns that name and append the index. 
	 * This method will use the activity's full name and strip any reserved character, 
	 * and make the name unique within one transformation session. 
	 * 
	 * @param activity	the activity that the name will be retrieved for or generated from. 
	 * @param index		the index that need to be appended to the name or null.
	 * @return String
	 */
	public  String getProcessName(Activity activity, Integer index){
			String name = null;
			String nameKey = getFullName(activity);
			int numberOfInputSet = activity.getInputParameterSet().size();
			int maxLength = 
					(numberOfInputSet > 1) ? 
					(MAX_PROCESS_NAME_LENGTH - String.valueOf(numberOfInputSet).length()) :
					 MAX_PROCESS_NAME_LENGTH;			
			if (index!=null){
				maxLength-=1+index.toString().length();
			}				
			
			if((name = (String)processNameRegistry.get(nameKey)) != null){
				if(name.length()> maxLength){
				   //trucate to make it in the range
				   name = name.substring(0,maxLength);	
				   name=name.trim();				
				}					
				name = (index == null)? name : (name + "_" + index.toString());
				return name;
			}
	
			//otherwise create one and register it

					
			name = activity.getName(); 
			name = stripChars(name, PROCESS_NAME_CHAR_NOT);
			if(name.length()> maxLength){
				//trucate to make it in the range
				name = name.substring(0, maxLength);	
				name=name.trim();				
			}	
		
			name= getUniqueName(name, 
						processNameRegistry.values(),
						maxLength);
		
				
			processNameRegistry.put(nameKey, name);
	
			name = (index == null)? name: name + "_" + index.toString();

			return name;
	
	}

	/**
	 * Returns process category name for the given activity.
	 * It searches the name registry first and if found a process categoty name already registered for the activity,
	 * it returns that name, otherwise, it generates a new one. 
	 * This method will use the activity's full package name and strip any reserved character, and shorten the length is necessary, 
	 * and make the name unique within one transformation session. 
	 * @param activity  the activity the process category name can be retrieved for or generated from. 
	 * @return String
	 */
	public  String getProcessCategoryName(Activity activity){
	
		String nameKey =null;
		if(activity.getOwningPackage()!= null){
			nameKey = getFullPackageName(activity.getOwningPackage()); 
		}else{
			nameKey = activity.getName();
		}
		String name;
		if((name =(String) processCategoryNameRegistry.get(nameKey)) != null){
			return name;
		}
	
		//else create one
		name = nameKey;
		if(name.length() > MAX_PROCESS_CATEGORY_NAME_LENGTH){
			name= name.substring(0, MAX_PROCESS_CATEGORY_NAME_LENGTH);
			name=name.trim();
		}
		name = getUniqueName(name, 
				processCategoryNameRegistry.values(),
				MAX_PROCESS_CATEGORY_NAME_LENGTH);
			
		processCategoryNameRegistry.put(nameKey, name);		
		return name;
	}


	/**
	 * Returns data structure name for the given pinSet.
	 * It searches the name registry first and if found a data structure name already registered for the pinSet,
	 * it returns that name, otherwise, it generates a new one. 
	 * This method will use the pinSet's full name and strip any reserved character, and shorten the length is necessary, 
	 * and make the name unique within one transformation session.
	 *  
	 * @param pinSet  the pinSet the data structure is retrieved for or generated from. 
	 * @return String
	 */
	public  String getDataStructureName(PinSet pinSet){
	
		List objPins;
		if (pinSet instanceof InputPinSet)
		{
			objPins = ((InputPinSet)pinSet).getInputObjectPin();		
		} else {
			objPins = ((OutputPinSet)pinSet).getOutputObjectPin();					
		}
		if (objPins.size()==0){
			return null;
		}	
	
		String nameKey = doComputeDataStructureKey(objPins);
		String name;
		if((name = (String)dataStructureNameRegistry.get(nameKey)) != null){
			return name;		
		}
	
		name = pinSet.getName(); 
		name = name.trim();
		//	TODO: although the MQ Workflow document FDL definition says the Datastructure name can be any character except control character,
		// but the MQ Buildtime doesn't allow some characters, so restrict it here  
		name = stripChars(name, DATA_STRUCTURE_NAME_CHAR_NOT);
	
		if(name.length()> MAX_DATA_STRUCTURE_NAME_LENGTH){
			//trucate to make it in the range
			name = name.substring(0,  MAX_DATA_STRUCTURE_NAME_LENGTH);		
			name=name.trim();			
		}	

		name= getUniqueName(name, 
					dataStructureNameRegistry.values(),
					MAX_DATA_STRUCTURE_NAME_LENGTH);
						
		dataStructureNameRegistry.put(nameKey, name);
		return name;
	}


	/**
	 * Returns data structure name for the given type.
	 * It searches the name registry first and if found a data structure name already registered for the type,
	 * it returns that name, otherwise, it generates a new one. 
	 * This method will use the type's full name and strip any reserved character, shorten the length is necessary, 
	 * and make the name unique within one transformation session.
	 *  
	 * @param type  the type the data structure is retrieved for or generated from. 
	 * @return String
	 */
	public  String getDataStructureName(Type type){
			String nameKey = getFullName(type);
			String name;
			if((name = (String)dataStructureNameRegistry.get(nameKey))!= null){
				return name;
			}
		
			//else create one
			name = getTypeName(type);
			//TODO: although the MQ Workflow document FDL definition says the Datastructure name can be any character except control character,
			// but the MQ Buildtime doesn't allow some characters, so restrict it here  
			name = stripChars(name, DATA_STRUCTURE_NAME_CHAR_NOT);
		
			if(name.length()> MAX_DATA_STRUCTURE_NAME_LENGTH){
				//trucate to make it in the range
				name = name.substring(0,  MAX_DATA_STRUCTURE_NAME_LENGTH);	
				name=name.trim();				
			}	

			name= getUniqueName(name, 
						dataStructureNameRegistry.values(),
						MAX_DATA_STRUCTURE_NAME_LENGTH);
						
			dataStructureNameRegistry.put(nameKey, name);
			return name;			
 
	}

	public String getUniqueProgramName(String pNameKey) {
		return DEFAULT_PROGRAM_NAME_PREFIX+"_"+getKeyFromString(pNameKey);
	}

	public String getUniqueDataStructureName(String dsNameKey) {
		return DEFAULT_DATA_STRUCTURE_NAME_PREFIX+"_"+getKeyFromString(dsNameKey);
	}
	
	public String getKeyFromString (String string) {
		int hashcode = string.hashCode();
		StringBuffer strBuf = new StringBuffer();
		String hashCodeString = Integer.toString(hashcode);
		if (hashcode < 0) {
			strBuf.append("0");
			strBuf.append(hashCodeString.substring(1));
		} else {
			strBuf.append(hashCodeString);
		}
		return strBuf.toString();		
	}
	
	/**
	 * Returns a key that is unique for the composition of pin types and multiplicity of the pins in the pinSet. 
	 * This key can be used to classify identical structure composition of different pinSets.
	 * @param pinSet	the PinSet that the key is to be computed from. 
	 * @return String 
	 */
	public  String computeDataStructureKey(PinSet pinSet){
		List objPins;
			if (pinSet instanceof InputPinSet)
			{
				objPins = ((InputPinSet)pinSet).getInputObjectPin();		
			} else {
				objPins = ((OutputPinSet)pinSet).getOutputObjectPin();					
			}
			if (objPins.size()==0){
				return null;
			}	
	
			return doComputeDataStructureKey(objPins);	
	}

	private String doComputeDataStructureKey(List objPins){
		ObjectPin objPin;
		StringBuffer strBuf = new StringBuffer();
		int multiplicity;
		for (int i=0; i<objPins.size(); i++)
		{
			objPin = (ObjectPin)objPins.get(i);
			multiplicity = getMultiplicity(objPin.getUpperBound());
			strBuf.append(
				((i>0)?"_":"")+
				getFullName(objPin.getType())+
				((multiplicity>1)?"_"+multiplicity:"")
			);
		}
		return strBuf.toString().trim()+"Container";
	}

	private int getMultiplicity(ValueSpecification lun)
	{
		// if no multiplicity specify, default to 1
		if (lun == null)
			return 1;

		String val;
		if (lun instanceof LiteralUnlimitedNatural)
			{
				val = ((LiteralUnlimitedNatural)lun).getValue();
			} else {
				val = ((LiteralInteger)lun).getValue().toString();
			}

		// if unbound or zero multiplicity is found, default to 1
		if (val.toLowerCase().equals("n") || val.equals("0") || val.equals("*"))
		{
			val="1";
		}
		return Integer.parseInt(val);
	}


	/**
	 * Returns the canonical name for the given NamedElement by prepending the name of
	 * the owner object up through the ownership hierarchy.  This name can be assumed
	 * to be unique within the given process only.  
	 * 
	 * Note: This version of the method does not include the containing Activity name
	 * nor the name of the containing top level SAN. 
	 * 
	 * Note also: This method is likely only useful to the TE, because it does not 
	 * guarantee that names are unique across processes since it does not include the name
	 * of the process model, or any parent process models, if nested. 
	 * 
	 * @param element  The <code>NamedElement</code> to generate the name for.
	 * @return String - The canonical name of <code>element</code>
	 */
	private String getCanonicalName(NamedElement element){
		return getCanonicalName(element, true, false, false);
	}
	
	
	private String getCanonicalName(NamedElement namedElement, boolean includeSelf, boolean includeActivityName, boolean includeTopLevelSANName){
		StringBuffer sb = new StringBuffer();	
		String name;
		if(includeSelf){
			sb.append(namedElement.getName());
		}
	
		NamedElement container = (NamedElement)namedElement.eContainer();
		if(container != null){
			if(container instanceof ProcessModel){
				return sb.toString();
			}else if(!includeActivityName && container instanceof Activity){
				//do nothing
			}else if(!includeTopLevelSANName &&  (container instanceof Action) 
			&& BomUtils.isTopLevelProcess((Action)container)){
				//do nothing
			}else{ 
				if(includeSelf){		
					sb.insert(0, container.getName()+ "_");
				}else{
					sb.insert(0, container.getName());
				}
			}
		}
		if(container != null){
			container = (NamedElement)container.eContainer();
			while (container != null && !(container instanceof ProcessModel)){
				if (!includeActivityName && (container instanceof Activity)){
					// do nothing
				} else if (!includeTopLevelSANName 
							&& (container instanceof Action) 
								&& BomUtils.isTopLevelProcess((Action)container)){
					//do nothing				
				}else{
					sb.insert(0, container.getName() + "_");
				}				
				container = (NamedElement)container.eContainer();
		
			}
		}
	
		name = sb.toString();
		if(includeSelf == false && name.equals("")){
			name = namedElement.getName();
		}
	
		return name;	
	
	}


	private String getUniqueName(String name, Collection nameRegistry, int maxLength){
	

		if(!nameRegistry.contains(name)){
			return name;
		}
		//else try to create one
		int suffixCounter = 1;
		
		String newName = name;
		while(nameRegistry.contains(newName)){
			String counterStr = String.valueOf(suffixCounter);
			int counterLength = counterStr.length();
			if((newName.length() + counterLength) <= maxLength){
				newName = name + counterStr;
			}else{
				newName = name.substring(0, maxLength - counterLength) + counterStr;
			} 
			suffixCounter++;
		}
	
		return newName.trim();
	
	}
	
	private String stripChars(String name, char[] strippedChars){
			StringTokenizer st = new StringTokenizer(name, new String(strippedChars));
			StringBuffer sb = new StringBuffer();
			while(st.hasMoreTokens()){
				sb.append(st.nextToken());
			}			
			return stripSpaces(sb.toString());		
		}
	
	private String stripSpaces(String name){
		StringTokenizer st = new StringTokenizer(name, " ");
		StringBuffer sb = new StringBuffer();
		while(st.hasMoreTokens()){
			String nextToken = st.nextToken();
			if (nextToken.length()!=0)
				sb.append(nextToken);
				sb.append(" ");
		}
		return 	sb.toString().trim();
	}
	
	private String stripWords(String name, String[] restrictedWords){	
	
		StringTokenizer st = new StringTokenizer(name);
		List wordsList = Arrays.asList(restrictedWords); 
		if(st.countTokens() > 1){
			StringBuffer sb = new StringBuffer();
			while(st.hasMoreElements()){
				String token = st.nextToken(); 
				if( wordsList.contains(token)){
					sb.append("_" + token);
				}else{
					sb.append(token);
					sb.append(" ");
				}
		
			}
		 name = sb.toString();	
		}
		return name.trim();
		
	}


	/**
	 * Returns s trancated to fit the length of FDL description restriction if necessary. 
	 */
	public String getDescription(String s){
		
		if(s == null)
			return null;
		s = stripChars(s, DESCRIPTION_CHAR_NOT);
		if(s.length() > MAX_DESCRIPTION_LENGTH){
			s = s.substring(0, MAX_DESCRIPTION_LENGTH);
			s=s.trim();
		}
		return s;
	}

	/**
	 *  Returns s trancated to fit the length of FDL documentation restriction if necessary.  	 
	 */
	public String getDocumentation(String s){
		if(s == null)
			return null;
		s = stripChars(s, DOCUMENTATION_CHAR_NOT);
		if(s.length() > MAX_DOCUMENTATION_LENGTH){
			s = s.substring(0, MAX_DOCUMENTATION_LENGTH);
			s=s.trim();
		}
		return s;
	}

	/**
	 *  Returns a string extracted and concatenated from the list. 
	 *  It may trancated to fit the length of FDL documentation restriction if necessary.
	 * 
	 * @param list a list of description strings
	 * @return String  
	 */
	public String getDocumentation(List list){
		if(list == null)
			return null;
		StringBuffer sb = new StringBuffer();	
		for(Iterator it =list.iterator(); it.hasNext();){
			sb.append(it.next());
			sb.append(" ");
		}
		String s = sb.toString();
		return getDocumentation(s);
	}
	public void resetForProcess(){
		activityNameRegistry.clear();
	}

	public void resetForSession(){
		processNameRegistry.clear();
		programNameRegistry.clear();
		dataStructureNameRegistry.clear();
		dataStructureMemberNameRegistry.clear();		
	}
	
			
	private String getFullName(PackageableElement pkgElement)
	{
		String name = null;
		// Ray: try to handle null packageElement
		if (pkgElement == null) {
			return "";
		}
		if (pkgElement instanceof Type) {
			Type type = (Type) pkgElement;
			name = getFullPackageName(type);
		} else {
			EObject container = pkgElement.eContainer();
			Activity belongingActivity = null;
			while (!(container instanceof Activity)) {
				container = container.eContainer();
			}
			name = pkgElement.getName()+"_"+getFullPackageName((Activity)container);
		}
		return name;
	}
	
	private String getFullName(NamedElement namedElement)
		{
			String name = null;
			name = namedElement.getName();
			NamedElement container = (NamedElement)namedElement.eContainer();
			if(container == null){
				return name;
			}
		
			while ((container != null) && !(container instanceof Activity)) {
				name = container.getName() + "_" + name;
				container = (NamedElement)container.eContainer();
			}
			if(container != null){
				name = getFullPackageName((Activity)container) + "_" + name;
			}
			
			return name;
		}


/**
 *  Returns the data structure memeber name for the objPin.
 *  This method first try to search if a name is already registered for th objPin within the ds, 
 *  if not found, it uses the objPin type's name and strip reserved characters and make it unique within the ds.
 * 
 * @param ds		the data structure that the objPin is part of.
 * @param objPin 	the object pin that the member name is retrieved for or generated from.
 * @return String
 */
	public String getDataStructureMemberName(DataStructure ds, ObjectPin objPin){
		String mbKey = getFullName(objPin);
		String dsKey = ds.getName() ;
		String name = getDataStructureMemberName(mbKey, dsKey);		
		if(name!= null){
			return name;
		}
		//else try to create name
		// Ray: Try to handle null type		
		if (objPin.getType() != null) {
			name = objPin.getType().getName();	
			name = name.trim();	
			if (name.length()>1) {
				name=name.substring(0,1).toLowerCase()+name.substring(1);
			} else {
				name=name.toLowerCase();
			}
		} else {
			name = "";
		}
		name = createDataStructureMemberName(mbKey, name, dsKey);
		return name;
	}
	
/**
 *  Returns the data structure memeber name for the property.
 *  This method first try to search if a name is already registered for th property within the datastructure that created out of the type, 
 *  if not found, it uses the property's name, strip reserved characters and make it unique within the datastructure.
 *  this method can be used when the property is from a super type and need to to be flatten into <code>type</code>.
 * 
 * @param property	the property that the name is retrieved for or generated from. 
 * @param type		the type that the datastructure is generated from in which the property is the member.
 * @return String 
 */
public String getDataStructureMemberNameForType(Property property, Type type){
		
			
			String mbKey = getFullName(property);
			String dsKey = getDataStructureName(type);
			String name = getDataStructureMemberName(mbKey, dsKey);
			if(name != null){
				return name;
			}
			
			//else try to create name
			name = property.getName().trim();
			name = createDataStructureMemberName(mbKey, name, dsKey);

			return name;
		
		
		}
		
/**
 *  Returns the data structure memeber name for the property.
 *  This method first try to search if a name is already registered for th property 
 *  within the datastructure that created out of the type that containing the property, 
 *  if not found, it uses the property's name, strip reserved characters and make it unique within the datastructure.
 * 
 * @param property	the property that the member name is retrieved for or generated from.
 * @return String
 */
public String getDataStructureMemberName(Property property){

					
		String mbKey = getFullName(property);
		String dsKey = getDataStructureName((Type)((NamedElement)property).eContainer());
		String name = getDataStructureMemberName(mbKey, dsKey);
		if(name!= null){
			return name;
		}
		
		//else try to create name
		name = property.getName();
		name=name.trim();
		name = createDataStructureMemberName(mbKey, name, dsKey);
		return name;

	}


private String getDataStructureMemberName(String memberKey, String dataStructureKey){
	
	Map mbRegistry = (Map)dataStructureMemberNameRegistry.get(dataStructureKey);
	if (mbRegistry == null){
		return null;
	}	
	
	return (String)mbRegistry.get(memberKey);				
}


private String createDataStructureMemberName(String memberKey, String proposedName, String dataStructureKey){
		
	String name = restrictDataStructureMemberName(proposedName);
	if(name.length() > MAX_DATA_STRUCTURE_MEMBER_NAME_LENGTH){
		name = name.substring(0, MAX_DATA_STRUCTURE_MEMBER_NAME_LENGTH);
		name = name.trim();
	}

	Map mbRegistry = (Map)dataStructureMemberNameRegistry.get(dataStructureKey);
	if (mbRegistry == null){
		mbRegistry = new HashMap();
		mbRegistry.put(memberKey, name);
		dataStructureMemberNameRegistry.put(dataStructureKey, mbRegistry);
	}
	else{
		name = getUniqueName(name, 
			mbRegistry.values(),  
			MAX_DATA_STRUCTURE_MEMBER_NAME_LENGTH);	
		mbRegistry.put(memberKey, name);		
	}

	return name;
}
	
private String getTypeName(Type type)
{
	String name=type.getName();
	name = name.trim();
// Ray: March 09, 2005 - Don't create the prefix for business item
//	Package belongingPackage = type.getOwningPackage();
//	while (belongingPackage != null && !ProcessUtil.isRootPackage(belongingPackage)) {
//		name = belongingPackage.getName().substring(0, 1) + "_" + name;
//		belongingPackage = belongingPackage.getOwningPackage();
//	}
	return name;
}


private String getFullPackageName(PackageableElement pkgElement)
{
	String name=pkgElement.getName();
	name=name.trim();
	Package belongingPackage = pkgElement.getOwningPackage();
	while (belongingPackage != null && !ProcessUtil.isRootPackage(belongingPackage)) {
		name = belongingPackage.getName() + "_" + name;
		belongingPackage = belongingPackage.getOwningPackage();
	}
	return name;
}
	

}
