package com.ibm.btools.te.ilm.heuristics.helper;
/*
* Licensed Materials - Property of IBM
* 5724-I74, 5724-I75
*  Copyright IBM Corporation 2003, 2009. All rights reserved.
* US Government Users Restricted Rights - Use, duplication or disclosure
* restricted by GSA ADP Schedule Contract with IBM Corp.
*/
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import com.ibm.btools.bom.model.artifacts.NamedElement;
import com.ibm.btools.bom.model.processes.activities.Action;
import com.ibm.btools.bom.model.processes.activities.InputPinSet;
import com.ibm.btools.bom.model.processes.activities.OutputPinSet;
import com.ibm.btools.bom.model.processes.activities.Pin;
import com.ibm.btools.bom.model.processes.activities.PinSet;

/**
 * This class defines methods that are used to find concurrent and exclusive merge points
 * in a BOM model instance, given their respective branch points.
 * 
 * Note: This class is only available to be used directly by other classes in this package.
 * All other classes that need to find a merge point should do so using the BomHelper class, since
 * it caches results and provides a general API.
 */
public class MergePointFinder {

	public static final String copyright = "Licensed Material - Property of IBM " + " 5724-I74, 5724-I75 " + "(C) Copyright IBM Corporation 2003, 2009. All Rights Reserved. " + "U.S. Government Users Restricted Rights - Use, duplication or disclosure " + "restricted by GSA ADP Schedule Contract with IBM Corp.".intern(); //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$ //$NON-NLS-4$ //$NON-NLS-5$
	
	/**
	 * The node path context that is used to find the branch point
	 */
	private NodePathContext fNodePathContext;
	
	/**
	 * Indicates whether a merge point has been found. This enables the getNodePathContext() method
	 * to return a node path context.
	 */
	private boolean fMergePointFound;
	
	/**
	 * The BOM Walker
	 */
	private BomWalker fBomWalker;
	
	/**
	 * The Default Constructor.
	 */
	public MergePointFinder(){
		fBomWalker = new BomWalker();
		fMergePointFound = false;
	}
	

	/**
	 * Given a particular branch point, finds the corresponding full concurrent merge point 
	 * 
	 * @param branchPoint The OutputPinSet that is a concurrent branch point
	 * @return NamedElement - The merge point which will be either an InputPinSet or a 
	 * StructuredActivityNode that is not a task
	 */
	public NamedElement findConcurrentMergePoint(OutputPinSet branchPoint){
		NamedElement result = null;
		fMergePointFound = false;

		/* Initialize the context for this merge point */
		fNodePathContext = new NodePathContext();

		/* Create a node path for this output set, and put it in the context */
		NodePath nodePath = new NodePath();
		nodePath.setElement( branchPoint );
		fNodePathContext.registerNodePath(nodePath);

		/* Start the long and winding road of recursion... */
		result = internalFindConcurrentMergePoint(nodePath, true);
		
		if (result != null){
			fMergePointFound = true;
		}
		
		return result;

	}

	/**
	 * Does the same thing as findConcurrentMergePoint, but recursively.
	 */
	private NamedElement internalFindConcurrentMergePoint(NodePath nodePath, boolean firstTime){
		
		NamedElement result = null;
		
		/* Get the list of input sets that we can reach from the input/output set associated with this nodepath */
		/* TODO: Should we use the jumping version of getNext(...) like the exclusive case (below)? */
		List possibleMergePoints = null;
		if (firstTime){
			possibleMergePoints = fBomWalker.getNext(nodePath.getElement(), BomWalker.INPUT_SET, false);
		} else {
			possibleMergePoints = fBomWalker.getNext(nodePath.getElement(), BomWalker.INPUT_SET, true);
		}
		//added additional check for special case of a concurrent point that branches to a SAN where the SAN is not returned as a possible
		//mergePoint.
		if(firstTime && nodePathConnectedToSAN(nodePath, possibleMergePoints)){
			//force to set the mergePoint to null, so the SAN will be returned.
			possibleMergePoints = null;
		}
		/* For each input set just returned (after the first one), make a clone the node path and
		 * update the element to be one of the input sets just returned. */
		if (possibleMergePoints != null 
				&& !possibleMergePoints.isEmpty()){

			List nodePaths = new ArrayList();
			Iterator it = possibleMergePoints.iterator();

			/* We will use the first input set in the current node path, after we've cloned it */
			InputPinSet firstInputPinSet = (InputPinSet)it.next();
			
			/* Clone the original node path... a new node path for each additional input set */
			NamedElement visitedInputPinSet = null;
			while (it.hasNext()){
				visitedInputPinSet = (InputPinSet)it.next();
				NodePath newNodePath = (NodePath)nodePath.clone();
				fNodePathContext.registerNodePath(newNodePath);

				newNodePath.setElement(visitedInputPinSet);
				newNodePath.recordVisited(visitedInputPinSet);
				fNodePathContext.updateNodePathRecord(newNodePath);

				nodePaths.add(newNodePath);
			}

			/* Reuse the current node path for the first input set */
			nodePath.recordVisited(firstInputPinSet);
			fNodePathContext.updateNodePathRecord(nodePath);

			nodePath.setElement(firstInputPinSet);
			nodePaths.add(nodePath);
		
			/* Now that the node paths are all created and registered, check if we have found
			 * a merge point.  If all the node paths in existence have now visited the same input set, 
			 * we are finished.
			 *   
			 * (NOTE: This is the base case for the recursion)	 */
			result = fNodePathContext.findMerge();

			/* If result == null, we haven't yet found a merge point, so start the process again 
			 * for each node path */
			for (it = nodePaths.iterator(); it.hasNext() && result == null; ){
				result = internalFindConcurrentMergePoint((NodePath)it.next(), false);
			}

		} else {
			/* There are three cases where possibleMergePoints is empty:
			 * 1. There are no output pins in the output set correlated with the input set we are starting from
			 * 2. All the output pins in the correlated output set attach to ControlNodes (or to
			 *    nothing at all), and not input pins on other actions
			 * 3. Getting to the next input set would involve stepping out of a SAN container
			 */
			 
			/* We can treat all these cases as if they are case 3, because the SAN container
			 * will be the last possible candidate merge point in each case.
			 */
			 
			/* Case 3: If we would have to step OUT of a SAN, that means that the input set we are starting
			 * from belongs to an action that is in the SAN that we would have to step out of.  
			 * Record that SAN in the nodepath context.
			 */
			Action san = null;
			if (nodePath.getElement() instanceof InputPinSet){
			 	san = (Action)((InputPinSet)nodePath.getElement()).getAction().eContainer();
			} else if (nodePath.getElement() instanceof OutputPinSet){
				san = (Action)((OutputPinSet)nodePath.getElement()).getAction().eContainer();
			}
			nodePath.recordVisited(san);
			fNodePathContext.updateNodePathRecord(nodePath);

			/* See if we have a merge point yet. Return the result because there is no where else to go! */
			result = fNodePathContext.findMerge();
			 
		}
		
		return result;
	}


	/**
	 * Returns the full exclusive merge point for the branch point given as an argument. 
	 * 
	 * A full exclusive merge point is defined as the OutputPinSet through which each of the original 
	 * paths eventually flows out of.  If there is no such OutputPinSet prior to the end of the SAN that
	 * contains the exclusive branch point, the merge point is defined as the SAN itself.
	 * 
	 * @param branchPoint - An InputPinSet that is a branch point
	 * @return NamedElement - The merge point that corresponds to the branch point.  Must be either an
	 * OutputPinSet or a StructuredActivityNode  
	 */
	public NamedElement findExclusiveMergePoint(InputPinSet branchPoint){
		NamedElement result = null;
		fMergePointFound = false;

		/* Initialize the context for this merge point */
		fNodePathContext = new NodePathContext();

		//List correlatedOutputPinSets = branchPoint.getOutputPinSet();
		List correlatedOutputPinSets = BomUtils.getCorrelatedPinSets(branchPoint);
		
		if (correlatedOutputPinSets.size() == 1){
			/* This is the trivial case where the branch point isn't even a branch point because
			 * the input set only correlated to one output set
			 */
			fMergePointFound = true;
			result = (NamedElement)correlatedOutputPinSets.get(0);
			
		} else {
		
			OutputPinSet correlatedOutputPinSet = null;
			Iterator it3 = correlatedOutputPinSets.iterator();
			List nodePathList = new ArrayList();
			while (it3.hasNext()){
				correlatedOutputPinSet = (OutputPinSet)it3.next();
				
				/* Create a node path for this output set and put it in the context, unless
				 * the output set doesn't have any output pins, in which case it would lead nowhere,
				 * and we would never find a merge point
				 */
//				if (correlatedOutputPinSet.getOutputObjectPin().size() > 0 
//					|| correlatedOutputPinSet.getOutputControlPin().size() > 0){
				//comment out to handle the empty outputset for which the merge point will not be counted - 08/19/2008
					NodePath nodePath = new NodePath();
					nodePathList.add(nodePath);
					
					/* Set the output set as the element of this node path */
					nodePath.setElement( correlatedOutputPinSet );
					nodePath.recordVisited(correlatedOutputPinSet);
					fNodePathContext.registerNodePath(nodePath);
//				}
			}

			NamedElement thisResult = null;
			for (Iterator it2 = nodePathList.iterator(); it2.hasNext(); ){
				/* Start the wild ride of recursion for each nodePath! */
				thisResult = internalFindExclusiveMergePoint((NodePath)it2.next());
				if (thisResult != null){
					fMergePointFound = true;
					return (NamedElement)thisResult;
				}
			}
		}
		
		if (result != null){
			fMergePointFound = true;
		}
		return result;
	}

	public NamedElement findExclusiveMergePoint(List<InputPinSet> branchPoints){
		NamedElement result = null;
		fMergePointFound = false;

		/* Initialize the context for this merge point */
		fNodePathContext = new NodePathContext();

			Iterator it3 = branchPoints.iterator();
			List nodePathList = new ArrayList();
			InputPinSet correlatedOutputPinSet;
			while (it3.hasNext()){
				correlatedOutputPinSet = (InputPinSet)it3.next();
				
				/* Create a node path for this output set and put it in the context, unless
				 * the output set doesn't have any output pins, in which case it would lead nowhere,
				 * and we would never find a merge point
				 */
				if (correlatedOutputPinSet.getInputObjectPin().size() > 0 
					|| correlatedOutputPinSet.getInputControlPin().size() > 0){

					NodePath nodePath = new NodePath();
					nodePathList.add(nodePath);
					
					/* Set the output set as the element of this node path */
					nodePath.setElement( correlatedOutputPinSet );
					nodePath.recordVisited(correlatedOutputPinSet);
					fNodePathContext.registerNodePath(nodePath);
				}
			}

			NamedElement thisResult = null;
			for (Iterator it2 = nodePathList.iterator(); it2.hasNext(); ){
				/* Start the wild ride of recursion for each nodePath! */
				thisResult = internalFindExclusiveMergePoint((NodePath)it2.next(),BomWalker.OUTPUT_SET_FROM_PROCESS_INPUTSET);
				if (thisResult != null){
					fMergePointFound = true;
					return (NamedElement)thisResult;
				}
			}
		
		
		if (result != null){
			fMergePointFound = true;
		}
		return result;
	}

	/**
	 * This internal method works recursively to step from output set to output set, recording
	 * each step along the way, and creating new nodePaths for each branch in the path.
	 * 
	 * @param nodePath The nodePath from which to start this leg of the journey
	 * @return NamedElement - The merge point, or null if one could not (yet) be found along the
	 * paths stemming from this nodePath.
	 */
	private NamedElement internalFindExclusiveMergePoint(NodePath nodePath){
		return internalFindExclusiveMergePoint(nodePath,BomWalker.OUTPUT_SET);
	}
	private NamedElement internalFindExclusiveMergePoint(NodePath nodePath, int destType){
		
		NamedElement result = null;
		
		/* Get the list of output sets that we can reach from the output set associated with this nodepath */
		// QUESTION: Should this use the jumping version BomWalker.getNext(...)?
		// NO! We can't because this is the method that is called to find the merge points!! (Circular dep.)
		// WAIT! We could recursively figure out branch/merge points, as long as we use different instances
		//       of this MergePointFinder for each... (made sure of this in BomHelper.findXXXMergePoint(...))
		// YES! We have to because if there is a nested SAN in the EBR, we want to walk over it, not into it.  If
		//      we walk into it, we will get caught at the 'back wall' of the SAN.  If all EB paths go through this same
		//      nested SAN, we will incorrectly find that as the EBP.
		//		Furthermore, we want to jump over any nested EBRs and CBRs as well
		//List possibleMergePoints = fBomWalker.getNext(nodePath.getElement(), BomWalker.OUTPUT_SET, true);
		List possibleMergePoints = fBomWalker.getNext(nodePath.getElement(), destType, true);
//		List possibleMergePoints = fBomWalker.getNext(nodePath.getElement(), BomWalker.OUTPUT_SET, false);

		/* For each output set just returned (after the first one), clone the node path and set
		 * the element to one of the output sets.
		 */
		if (possibleMergePoints != null 
				&& !possibleMergePoints.isEmpty() 
				&& possibleMergePoints.get(0) instanceof OutputPinSet){

			List nodePaths = new ArrayList();
			Iterator it = possibleMergePoints.iterator();

			/* We will use the first output set in the current node path, after we've cloned it */
			OutputPinSet firstOutputPinSet = (OutputPinSet)it.next();
			
			/* Clone the original node path... a new node path for each additional output set */
			NamedElement visitedOutputPinSet = null;
			while (it.hasNext()){
				visitedOutputPinSet = (OutputPinSet)it.next();
				NodePath newNodePath = (NodePath)nodePath.clone();
				fNodePathContext.registerNodePath(newNodePath);

				newNodePath.setElement(visitedOutputPinSet);
				newNodePath.recordVisited(visitedOutputPinSet);
				fNodePathContext.updateNodePathRecord(newNodePath);

				nodePaths.add(newNodePath);

			}

			/* Reuse the current node path for the first output set */
			nodePath.recordVisited(firstOutputPinSet);
			fNodePathContext.updateNodePathRecord(nodePath);

			nodePath.setElement(firstOutputPinSet);
			nodePaths.add(nodePath);
		
			/* Now that the node paths are all created and registered, check if we have found
			 * a merge point.  If all the node paths in existence have now visited the same output set, 
			 * we are finished.
			 *   
			 * (NOTE: This is the base case for the recursion)
			 */
			result = fNodePathContext.findMerge();

			/* If result == null, we haven't yet found a merge point, so start the process again 
			 * for each node path
			 */
			for (it = nodePaths.iterator(); it.hasNext() && result == null; ){
				result = internalFindExclusiveMergePoint((NodePath)it.next());
			}

		} else {
			/* There are three cases where possibleMergePoints is empty:
			 * 1. There are no output pins in the output set that we are starting from
			 * 2. All the output pins in correlated output set attach to ControlNodes (or to nothing
			 *    at all), and not input pins on other actions. 
			 * 3. Getting to the next output set would involve stepping out of a SAN container.  
			 *    Note: the output sets on a nested subprocess are condered to be outside the SAN container.  
			 */

			/* We can treat all these cases as if they are case 3, because the SAN container
			 * will be the last possible candidate merge point in each case.
			 */
			 				 
			/* Case 3: If we would have to step OUT of a SAN, that means that the output set we are starting
			 * from belongs to an action that is in the SAN that we would have to step out of.  
			 * Record that SAN in the nodepath context.
			 */
			Action san = null;
			if (nodePath.getElement() instanceof OutputPinSet){
				san = (Action)((OutputPinSet)nodePath.getElement()).getAction().eContainer();
				nodePath.recordVisited(san);
				fNodePathContext.updateNodePathRecord(nodePath);
			} 
	
			/* See if we have a merge point yet. If not, return because there is no where else to go! */
			result = fNodePathContext.findMerge();
		}		 
		
		return result;
	}
	
	/**
	 * Returns the node path context used to find the last requested merge point. Note that
	 * this method should only be called following a call to findXXXMergePoint(), otherwise
	 * it will return null.
	 * 
	 * @return NodePathContext - The NodePathContext constructed during the previous search for a
	 * merge point.  This context can be used to answer queries regarding which elements were
	 * in the merge region found.
	 */
	public NodePathContext getNodePathContext(){
		return (fMergePointFound 
				? fNodePathContext 
				: null);
	}
	/**
	 * This is to treat the special case of a branch point with 2 branches where one connects to a SAN. The BomWalker will return
	 * when first execute for the first time (jump = true) (a single inputSet) which will then only register the SAN as the mergePoint 
	 * but will not recognize it as mergePoint.
	 * If there are more then 1 possibleMergePoints, SAN will be recognized as possible merge point and returned as expected.
	 * Returns true if the number or Pins in the nodePath is bigger than 1 and connect to PinSets other then the mergePoint.. They can only connect
	 * in that case to the SAN.
	 * @param nodePath
	 * @param possibleMergePoints
	 * @return
	 */
	private boolean nodePathConnectedToSAN(NodePath nodePath, List possibleMergePoints){
		//defect 167194
		if(possibleMergePoints != null && possibleMergePoints.size() == 1){
		//if(possibleMergePoints != null && possibleMergePoints.size() > 0){
			if(nodePath.getElement() instanceof PinSet){
				PinSet pinSet = (PinSet)nodePath.getElement();
				List pins = BomUtils.getPinsInSet(pinSet);
				if(pins.size()> possibleMergePoints.size()){
					PinSet mergePoint = (PinSet)possibleMergePoints.get(0);	
					//check if all connected pins connected to the mergePoint.. 
					Iterator it = pins.iterator(); 
					while(it.hasNext()){
						Pin thisPin = (Pin)it.next(); 
						if(thisPin.getOutgoing() != null && thisPin.getOutgoing().getTarget() instanceof Pin){
							if (!findPinSetForPin((Pin)thisPin.getOutgoing().getTarget()).equals(mergePoint)){
								return true;
							}
						}
					}
				}
			}
		//AR - defect 167194, 90414
		//this is to handle the case where there are more than 2 branches where 1 connects to the SAN. This will fail to check, 
		//as possible mergePoints size will be >1. The check looks for the existance of a Pin connected to a Process or Subprocess. 
		}else if(possibleMergePoints.size() > 1){
			if(nodePath.getElement() instanceof PinSet){
				PinSet pinSet = (PinSet)nodePath.getElement();
				List pins = BomUtils.getPinsInSet(pinSet);
				if(pins.size()> possibleMergePoints.size()){
					Iterator it = pins.iterator(); 
					while(it.hasNext()){
						Pin thisPin = (Pin)it.next(); 
						if(thisPin.getOutgoing() != null && 
								thisPin.getOutgoing().getTarget() instanceof Pin && 
								BomUtils.findPinSetForPin((Pin)thisPin.getOutgoing().getTarget()) instanceof OutputPinSet
								&& (BomUtils.isTopLevelProcess(((OutputPinSet)BomUtils.findPinSetForPin((Pin)thisPin.getOutgoing().getTarget())).getAction()) ||
										BomUtils.isLocalProcess(((OutputPinSet)BomUtils.findPinSetForPin((Pin)thisPin.getOutgoing().getTarget())).getAction()))){
							return true;
						}
					}				
				}
			}
			
			
		}
			
		return false;
	}

	/**
	 * This method attempts to find the first pinSet (input or output) that contains a Pin.  Notice a pin can never appear
	 * in both input and output pinset.
	 * @param pin
	 * @return
	 */
	public PinSet findPinSetForPin(Pin pin) {
		Action action = (Action)pin.eContainer();
		Iterator it = action.getInputPinSet().iterator();
		InputPinSet inPinSet;
		while (it.hasNext()) {
			inPinSet = (InputPinSet)it.next();
			// it can never be control pin
			if (inPinSet.getInputObjectPin().contains(pin)) {
				return inPinSet;
			}else if(inPinSet.getInputControlPin().contains(pin)){
				return inPinSet;
			}
			
		}
		it = action.getOutputPinSet().iterator();
		OutputPinSet outPinSet;
		while (it.hasNext()) {
			outPinSet = (OutputPinSet)it.next();
			// it can never be control pin
			if (outPinSet.getOutputObjectPin().contains(pin)) {
				return outPinSet;
			}else if (outPinSet.getOutputControlPin().contains(pin)) {
				return outPinSet;
			}
		}
		return null;  // This should never happened!!
	}
}
