package com.ibm.btools.te.ilm.heuristics.naming;
/*
* Licensed Materials - Property of IBM
* 5724-I74, 5724-I75
*  Copyright IBM Corporation 2003, 2009. All rights reserved.
* US Government Users Restricted Rights - Use, duplication or disclosure
* restricted by GSA ADP Schedule Contract with IBM Corp.
*/
import org.eclipse.emf.common.command.Command;

import com.ibm.btools.bom.model.artifacts.NamedElement;
import com.ibm.btools.bom.model.processes.activities.Action;
import com.ibm.btools.bom.model.processes.activities.InputPinSet;
import com.ibm.btools.bom.model.processes.activities.OutputPinSet;
import com.ibm.btools.bom.model.processes.activities.StructuredActivityNode;
import com.ibm.btools.bom.model.processes.businessrules.BusinessRuleAction;
import com.ibm.btools.bom.model.processes.humantasks.HumanTask;
import com.ibm.btools.te.attributes.helper.AttributeNameConstants;
import com.ibm.btools.te.attributes.helper.TechnicalAttributesHelper;
import com.ibm.btools.te.attributes.helper.UpdateCommandHelper;
import com.ibm.btools.te.attributes.helper.ValidationRecordHelper;
import com.ibm.btools.te.framework.TransformationContext;
import com.ibm.btools.te.framework.TransformationRule;
import com.ibm.btools.te.ilm.ExportOperation;
import com.ibm.btools.te.ilm.LoggingUtil;
import com.ibm.btools.te.ilm.TransformationEngine;
import com.ibm.btools.te.ilm.heuristics.helper.BomUtils;
import com.ibm.btools.te.ilm.heuristics.helper.BpelOptionsUtil;
import com.ibm.btools.te.ilm.heuristics.helper.ProcessUtil;
import com.ibm.btools.te.ilm.heuristics.helper.SAUtil;
import com.ibm.btools.te.ilm.heuristics.helper.TransformationSessionUtil;
import com.ibm.btools.te.ilm.resource.MessageKeys;
import com.ibm.btools.te.validation.model.TeValidationRecord;
import com.ibm.websphere.sca.scdl.Component;

/**
* Generic NameProvider responsible for generating unique valid names. The generated names are sensitive 
* to the mode the Transformation is performed. 
* Class extending the NameProvider should provide implementation for generation of the default Name. 
*/
public abstract class NameProvider {

	public static final String copyright = "Licensed Material - Property of IBM " + " 5724-I74, 5724-I75 " + "(C) Copyright IBM Corporation 2008, 2009. All Rights Reserved. " + "U.S. Government Users Restricted Rights - Use, duplication or disclosure " + "restricted by GSA ADP Schedule Contract with IBM Corp.".intern(); //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$ //$NON-NLS-4$ //$NON-NLS-5$

	//the Transformation Context
	protected TransformationContext context;

	/**
	 * Generic NameValidator. Classes extending the NameProvider should implement their own validator.
	 */
	public class NameValidator{
		protected String validName;
		protected String errorCode;
		public boolean validateName(Object type, String name, NamingRegistry registry){
			return validateName(type, name, registry, null);
		}
		public boolean validateName(Object type, String name, NamingRegistry registry, Object parentContex){
			    JavaNCNameConverter converter = new JavaNCNameConverter();
			    String converetedName = converter.convertName(name);
			    
			    if(!converetedName.equals(name)){
			   		String params[]={name};
					LoggingUtil.logError(MessageKeys.TE_ATTRIBUTES_INVALID_NCNAME, params, null);
			   // 	System.out.println("The " + name + " is an invalid NCName");
			    }
			    validName = converter.convertName(registry, type, name, parentContext);
			    if(validName.equals(converetedName)){
			    	return true;
			    }else{
			    	String params[]={name};
//			    	PMR54206,999,161
			    	//Do not validate the component duplicate name here, we should do that in the end
			    	if(!(type instanceof Component)){
			    		LoggingUtil.logError(MessageKeys.TE_ATTRIBUTES_DUPLICATE_NAME, params, null);
			    	}
			    // 	System.out.println("The " + name + " is duplicate name");
			    }
			    return false;
			}
		
		public String getValidName(){
			return validName;
		}
		public String getErrorCode(){
			return errorCode;
		}
	}
	//validator for the class. Implementors for this class can subclass the NameValidator if additional functionality
	//is needed.
	protected NameValidator nameValidator; 
	protected Object parentContext;

	/**
	 * Generates an UNIQUE, VALID value for the <code>propertyAttribute</code> of the <code>type</code>
	 * @param forType  a BPEL or WSDL Object
	 * @param propertyName name of the attribute property
	 * @param fromSource  BOM source object
	 * @param registry a NamingRegistry  
	 * @return String default name.
	 */
	public abstract String getDefaultName(Object forType, String propertyName, NamedElement fromSource, NamingRegistry registry, Object parentContext);

	
	/**
	 * Template method that is responsible for generating valid names based on the mode of transformation. 
	 * @param forType a BPEL or WSDL Object
	 * @param fromSource BOM source object
	 * @param registry a NamingRegistry
	 * @param propertyName the name of the technical Attribute
	 * @return
	 */
	public String getName(Object forType, NamedElement fromSource, NamingRegistry registry, String propertyName){
		if(forType instanceof Class && registry == null){
			//this is an invalid use of the API, will return the default name of the source 
			return getName(fromSource);
		}
		//here is the logic of retrieving the name of the type...
		String name = null;
		boolean isCBA = isFromCBA(fromSource);
		if(BpelOptionsUtil.generateDefaultValuesWithOverwrite() &&
				!AttributeNameConstants.BUSINESS_RULE_NAME.equals(propertyName)){	// skip name generation for Business Rules
			//1.just generate default values and ignore the technical attributes
			name = getDefaultName(forType, propertyName, fromSource, registry, parentContext);
			if (propertyName.equals("#serviceComponent.name") && forType instanceof Component &&
					name.indexOf(ExportOperation.FILE_SEPARATOR) != -1) {
				name = NamingUtil.getNameForArtifact(name);
			}
			//2c. call the commandUpdateHelper to update the BOM Model TA with new value
			Command command = UpdateCommandHelper.createUpdateCommand((NamedElement)fromSource, propertyName, name);
			//2d. call the Registry Helper to add the command to the list of Commands
			if(isCBA){
				ProcessUtil.putUpdateCommandsInContext(context, command);
			}else{
				TransformationSessionUtil.appendAndExecuteUpdateCommand(TransformationEngine.getTransformationSession(), command);
			}
		}else if(BpelOptionsUtil.generateDefaultValuesNoOverwrite() &&
				!AttributeNameConstants.BUSINESS_RULE_NAME.equals(propertyName)){	// skip name generation for Business Rules
			//1. retrieve the technical attribute for this type... 
			String taName = (String)TechnicalAttributesHelper.getAttributeValue(fromSource, propertyName);
			//2. check if TA was set.. not null
			String validName = null;
			//3. validate if the value is valid
			if(taName != null && !taName.trim().equals("")){
				boolean isValid = nameValidator.validateName(forType, taName, registry);
				// 3a if valid, return the value
				if(isValid){
					name = taName;
				// 3b if not valid, correct the value and return the new value.. transformation should continue with the new corrected value
				}else{
					validName = nameValidator.getValidName();
					name = validName;
				}
				//2b if TA was not set, generate a default name
			}else{
				name = getDefaultName(forType, propertyName, fromSource, registry, parentContext);
				if (propertyName.equals("#serviceComponent.name") && forType instanceof Component &&
						name.indexOf(ExportOperation.FILE_SEPARATOR) != -1) {
					name = NamingUtil.getNameForArtifact(name);
				}
				//2c. call the commandUpdateHelper to update the BOM Model TA with new value
				Command command = UpdateCommandHelper.createUpdateCommand((NamedElement)fromSource, propertyName, name);
				//2d. call the Registry Helper to add the command to the list of Commands
				if(isCBA){
					ProcessUtil.putUpdateCommandsInContext(context, command);
				}else{
					TransformationSessionUtil.appendAndExecuteUpdateCommand(TransformationEngine.getTransformationSession(), command);
				}
			}
		}else if(BpelOptionsUtil.runTransformation() ||
				AttributeNameConstants.BUSINESS_RULE_NAME.equals(propertyName)){	// use this for name generation for Business Rules){
			//this means we are running the transformation with user setting without validating.. 
			//1. retrieve the technical attribute for this type... 
			String taName = (String)TechnicalAttributesHelper.getAttributeValue((NamedElement)fromSource, propertyName);
			//2. check if TA was set.. not null
			//3a if TA was set, return TA
			if(taName != null && !taName.trim().equals("")){
				boolean isValid = nameValidator.validateName(forType, taName, registry, parentContext);
				if(!isValid){
					//LoggingUtil.logError(MessageKeys.TE_ATTRIBUTES_INVALID_NAME);
					registry.deregister(forType, nameValidator.getValidName());
				}
				//for component name use registry
				if (fromSource instanceof Action && AttributeNameConstants.COMPONENT_NAME.equals(propertyName)){
//					name = NamingUtil.getNameForArtifact(NamingUtil.convertNameForAction((Action)fromSource, context, taName));
					name = NamingUtil.convertNameForAction((Action)fromSource, context, taName);
				} else {
					name = taName;
				}
				//we should register the name in the registry
//				registry.register(forType, name);
				registry.register(forType, NamingUtil.getNameForArtifact(name));
				//3b if TA not set, generate default value
			}else{	
				name = getDefaultName(forType, propertyName, fromSource, registry, parentContext);
			}
		}else if(BpelOptionsUtil.runValidation()){
			//1. retrieve the technical attribute for this type... 
			String taName = (String)TechnicalAttributesHelper.getAttributeValue(fromSource, propertyName);
			//2. check if TA was set.. not null
			String validName = null;
			if(taName != null && !taName.equals("")){
				//validName = validateName(forType, taName, context);
				boolean isValid = nameValidator.validateName(forType, taName, registry, parentContext);
				validName = nameValidator.getValidName();
				
				// 3a if valid, return the value
				if(isValid){
					name = taName;
				// 3b if not valid, correct the value and return the new value.. 
				}else{
					String errorCode = nameValidator.getErrorCode();
					//3c. call the ValidationRecoredHeler to create a validationRecord 
					TeValidationRecord validationRecord = ValidationRecordHelper.createRecord(fromSource, validName, null, propertyName, errorCode );
					//3d. call the RegistryHelper to update the listOfValidationRecord
					TransformationSessionUtil.addValidationRecord(TransformationEngine.getTransformationSession(), validationRecord);
					name = validName;
				}
			}else{
				//2b. if TA was not set, generate a default name
				name = getDefaultName(forType, propertyName, fromSource, registry, parentContext);
				if (propertyName.equals("#serviceComponent.name") && forType instanceof Component &&
						name.indexOf(ExportOperation.FILE_SEPARATOR) != -1) {
					name = NamingUtil.getNameForArtifact(name);
				}
			}
		}
		return name;
	}

	/**
	 * Generic method for retrieving the name, that is not sensitive to the Transformation running style. 
	 * @param forType a BPEL or WSDL Object
	 * @param fromSource BOM source object
	 * @param registry a NamingRegistry
	 * @return string default name 
	 */
	public String getName(Object forType, NamedElement fromSource, NamingRegistry registry){
		return getDefaultName(forType, null, fromSource, registry, parentContext);
	}
	
	/**
	 * Returns the name of the <code>fromSource</code>
	 * @param fromSource a BOM NamedElement
	 * @return the name value of the <code>fromSource</code>
	 */
	protected String getName(NamedElement fromSource){
		return fromSource.getName();
	}
	/**
	 * @return Returns the context.
	 */
	public TransformationContext getContext() {
		return context;
	}
	/**
	 * @param context The context to set.
	 */
	public void setContext(TransformationContext context) {
		this.context = context;
	}
	
	/**
	 * Returns the object parent context. 
	 * @return
	 */
	public Object getParentContex(){
		return parentContext;
	}
	/**
	 * Sets the object parent context
	 * @param parentContext
	 */
	public void setParentContext(Object parentContext){
		this.parentContext = parentContext;
	}
	
	/**
	 * Returns true if the namedElement, an Input/OutputPinSet belong to an Action which is called
	 * from the main Process through a CBA
	 * @param namedElement
	 * @return
	 */
	private boolean isFromCBA(NamedElement namedElement){
		Action action = null;
		if(namedElement instanceof InputPinSet){
			action = ((InputPinSet)namedElement).getAction();
		}else if (namedElement instanceof OutputPinSet){
			action = ((OutputPinSet)namedElement).getAction();
		}else if(namedElement instanceof StructuredActivityNode){
			action = (StructuredActivityNode)namedElement;
		}
		// we are only interested in the cases of PinSet or StructuredActivityNode
		if(action == null){
			return false;
		}
		Action thisProcess = null;
		if(getContext() == null){
			return false;
		}
		TransformationRule procDefRule = ProcessUtil.getProcessDefinitionRule(getContext());
		if(procDefRule != null){
			thisProcess = ((com.ibm.btools.bom.model.processes.activities.Activity)ProcessUtil.getProcessDefinitionRule(getContext()).getSource().get(0)).getImplementation();
		}
		if(thisProcess == null){
			return false;
		}
		if((action instanceof StructuredActivityNode) && ((StructuredActivityNode)action).getInStructuredNode() == null && 
				("SERVICE".equals(((StructuredActivityNode)action).getAspect() )|| 
						"TASK".equals(((StructuredActivityNode)action).getAspect()))){
			return true;
		//check if this is a remote process
		}else if(BomUtils.isTopLevelProcess(action) && thisProcess != action ){
			return true;
		}else{
			return false;
		}
		
	}
}
