package com.ibm.btools.te.ilm.heuristics.naming;
/*
* Licensed Materials - Property of IBM
* 5724-I74, 5724-I75
*  Copyright IBM Corporation 2003, 2009. All rights reserved.
* US Government Users Restricted Rights - Use, duplication or disclosure
* restricted by GSA ADP Schedule Contract with IBM Corp.
*/
import java.text.NumberFormat;
import java.text.ParseException;
import java.util.HashMap;

import org.eclipse.emf.ecore.EObject;

import com.ibm.btools.bom.model.artifacts.NamedElement;
import com.ibm.btools.bom.model.externalmodels.ServiceInterface;
import com.ibm.btools.bom.model.processes.actions.CallBehaviorAction;
import com.ibm.btools.bom.model.processes.activities.Action;
import com.ibm.btools.bom.model.processes.activities.Activity;
import com.ibm.btools.bom.model.processes.activities.PinSet;
import com.ibm.btools.bom.model.processes.activities.StructuredActivityNode;
import com.ibm.btools.bom.model.time.RecurringTimeIntervals;
import com.ibm.btools.te.framework.TransformationContext;
import com.ibm.btools.te.framework.TransformationRule;
import com.ibm.btools.te.framework.impl.TransformationRuleImpl;
import com.ibm.btools.te.ilm.CrossProjectReferenceUtil;
import com.ibm.btools.te.ilm.ExportOperation;
import com.ibm.btools.te.ilm.heuristics.abstractbpel.DatastoreRule;
import com.ibm.btools.te.ilm.heuristics.abstractbpel.ProcessDefinitionRule;
import com.ibm.btools.te.ilm.heuristics.abstractbpel.ProcessInterfaceRule;
import com.ibm.btools.te.ilm.heuristics.abstractbpel.util.TELNamingRegistry;
import com.ibm.btools.te.ilm.heuristics.br.util.BRNamingRegistry;
import com.ibm.btools.te.ilm.heuristics.helper.BomUtils;
import com.ibm.btools.te.ilm.heuristics.helper.BpelOptionsUtil;
import com.ibm.btools.te.ilm.heuristics.helper.SAUtil;
import com.ibm.btools.te.ilm.heuristics.scdl.util.ScdlNamingRegistry;
import com.ibm.btools.te.ilm.heuristics.xsd.DataDefinitionRule;
import com.ibm.wbit.bpel.Flow;

public class NamingUtil {
	
	public static final String copyright = "Licensed Material - Property of IBM " + " 5724-I74, 5724-I75 " + "(C) Copyright IBM Corporation 2003, 2009. All Rights Reserved. " + "U.S. Government Users Restricted Rights - Use, duplication or disclosure " + "restricted by GSA ADP Schedule Contract with IBM Corp.".intern(); //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$ //$NON-NLS-4$ //$NON-NLS-5$
	public static String NAMING_REGISTRIES_SUB_CONTEXT = "NamingRegistriesSubContext";
	public static String DEFAULT_REGISTRY_SCDL="SCDL_DEFAULT_REGISTRY";
	public static String BR_NAMING_REGISTRY="BR_NAMING_REGISTRY";
	public static String TEL_NAMING_REGISTRY="TEL_NAMING_REGISTRY";
	public static String NO_HASH_NAMING_REGISTRY = "NO_HASH_NAMING_REGISTRY";
	public static String PROJECT_NAMING_REGISTRY = "PROJECT_NAMING_REGISTRY";
	public static String GLOBAL_XPROJECT_UNIQUENESS = "GLOBAL_XPROJECT_UNIQUENESS";
	public static boolean USE_PER_PROJECT_UNIQUENESS = false;
	
	/**	
	 * Returns the naming registry from the context 
	 * @param context  transformation context
	 * @param key  EObject key needed to retrieve the registry	 
	 * @return NamingRegistry the NamingRegistry from the context
	 */	
	static public NamingRegistry getRegistry(TransformationContext context, EObject key)
	{
		HashMap namingRegistriesSubContext;
		namingRegistriesSubContext = (HashMap)context.get(NAMING_REGISTRIES_SUB_CONTEXT);
		if (namingRegistriesSubContext != null) {
			return (NamingRegistry)namingRegistriesSubContext.get(key);
		}
		else
			return null;
	}

	
	/**	
	 * It puts the Registry into the transformation context.	 
	 * @param context  the transformation context
	 * @param key  EObject acts as the key to the NamingRegistriesSubContext, could be a bpel
	 *             process, WSDL object, etc.
	 * @param registry  NamingRegistry to be put into the context	 
	 */	
	static public void putRegistry(TransformationContext context, Object key, NamingRegistry registry)
	{
		HashMap namingRegistriesSubContext;
		namingRegistriesSubContext = (HashMap)context.get(NAMING_REGISTRIES_SUB_CONTEXT);
		if (namingRegistriesSubContext == null) {
			namingRegistriesSubContext = new HashMap();
			context.put(NAMING_REGISTRIES_SUB_CONTEXT,namingRegistriesSubContext );
		}
		namingRegistriesSubContext.put(key,registry);
	}

	/**
	 * Returns a NamingRegistry corresponding to the specific rule
	 * Checks the eContainers to find the appropriate EObject as the key to the registry	
	 * @param rule the transformation rule calling this class
	 * @return NamingRegistry the corresponding registry
	 */			
	static public NamingRegistry findRegistry(TransformationRule rule)
	{
		EObject containerObj = (EObject)rule;
		while (containerObj!=null)
		{
			if (    (containerObj instanceof ProcessDefinitionRule) || 
					(containerObj instanceof ProcessInterfaceRule) || 
					(containerObj instanceof DataDefinitionRule) ||								
					(containerObj instanceof DatastoreRule)){
						if(((TransformationRule)containerObj).getTarget()!= null && !((TransformationRule)containerObj).getTarget().isEmpty()){
							return NamingUtil.getRegistry(((TransformationRuleImpl)rule).getContext(), (EObject)((TransformationRule)containerObj).getTarget().get(0));			
						}else{
							return null;
						}
			}			
			else {
				containerObj = containerObj.eContainer();
			}
		}		
		return null;
	}
	
	/**
	 * Return the global default registry with the specific registryType:
	 * 1. SCDL_DEFAULT_REGISTRY
	 * 2.
	 * 3.
	 * @param context
	 * @param registryType
	 * @return
	 */
	public static NamingRegistry getDefaultRegistry (TransformationContext context,String registryType) {
		
		HashMap namingRegistriesSubContext;
		namingRegistriesSubContext = (HashMap)context.get(NAMING_REGISTRIES_SUB_CONTEXT);
		if (namingRegistriesSubContext != null) {
			if (namingRegistriesSubContext.get(registryType)==null) {
				if (registryType.equals(DEFAULT_REGISTRY_SCDL)) {
					namingRegistriesSubContext.put(registryType,new ScdlNamingRegistry());
				}
			}
			return (NamingRegistry)namingRegistriesSubContext.get(registryType);				
		} else {
			//we need to create a registry for the external service
			namingRegistriesSubContext = new HashMap();
			context.put(NAMING_REGISTRIES_SUB_CONTEXT,namingRegistriesSubContext );
			if (namingRegistriesSubContext.get(registryType)==null) {
				if (registryType.equals(DEFAULT_REGISTRY_SCDL)) {
					namingRegistriesSubContext.put(registryType,new ScdlNamingRegistry());
				}
			}
			return (NamingRegistry)namingRegistriesSubContext.get(registryType);
			//return null;			
		}		
	}
	
	/**
	 * Return the global BR registry with the specific registryType:
	 * 1. BR_NAMING_REGISTRY
	 * @param context
	 * @param registryType
	 * @return
	 */
	public static NamingRegistry getBRNamingRegistry (TransformationContext context,String registryType) {
		
		HashMap namingRegistriesSubContext;
		namingRegistriesSubContext = (HashMap)context.get(NAMING_REGISTRIES_SUB_CONTEXT);
		if (namingRegistriesSubContext != null) {
			if (namingRegistriesSubContext.get(registryType)==null) {
				if (registryType.equals(BR_NAMING_REGISTRY)) {
					namingRegistriesSubContext.put(registryType,new BRNamingRegistry());
				}
			}
			return (NamingRegistry)namingRegistriesSubContext.get(registryType);				
		} else {
			return null;
			
		}		
	}
	
	/**
	 * Return the global BR registry with the specific registryType:
	 * 1. BR_NAMING_REGISTRY
	 * @param context
	 * @param registryType
	 * @return
	 */
	public static NamingRegistry getTELNamingRegistry (TransformationContext context) {
		
		HashMap namingRegistriesSubContext;
		namingRegistriesSubContext = (HashMap)context.get(NAMING_REGISTRIES_SUB_CONTEXT);
		NamingRegistry nr = null;
		if (namingRegistriesSubContext == null) {
			if (namingRegistriesSubContext == null) {
				namingRegistriesSubContext = new HashMap();
				context.put(NAMING_REGISTRIES_SUB_CONTEXT,namingRegistriesSubContext );
			}
		}
		nr = (NamingRegistry)namingRegistriesSubContext.get(TEL_NAMING_REGISTRY);
		if (nr==null) {
			nr = new TELNamingRegistry();
			namingRegistriesSubContext.put(TEL_NAMING_REGISTRY,nr);
		}
		return nr;				
	}

	/**
	 * Return the global no hash registry with the specific registryType:
	 * 1. NO_HASH_NAMING_REGISTRY
	 * @param context
	 * @param registryType
	 * @return
	 */
	public static NamingRegistry getNoHashNamingRegistry (TransformationContext context,String registryType) {	
		HashMap namingRegistriesSubContext;
		namingRegistriesSubContext = (HashMap)context.get(NAMING_REGISTRIES_SUB_CONTEXT);
		if (namingRegistriesSubContext != null) {
			if (namingRegistriesSubContext.get(registryType)==null) {
				if (registryType.equals(NO_HASH_NAMING_REGISTRY)) {
					namingRegistriesSubContext.put(registryType, new NoHashNamingRegistry());
				}
			}
			return (NamingRegistry)namingRegistriesSubContext.get(registryType);				
		} else {
			return null;		
		}		
	}

	
	/**
	 * Return the global no hash registry with the specific registryType:
	 * 1. PROJECT_NAMING_REGISTRY
	 * @param context
	 * @param registryType
	 * @return
	 */
	public static NamingRegistry getProjectNamingRegistry (TransformationContext context, String registryType) {
		if (!registryType.equals(PROJECT_NAMING_REGISTRY)) {
			return null;
		}
		
		HashMap namingRegistriesSubContext;
		namingRegistriesSubContext = (HashMap)context.get(NAMING_REGISTRIES_SUB_CONTEXT);
		NamingRegistry projectRegistry = null;
		if (namingRegistriesSubContext == null) {
			projectRegistry = new ProjectNamingRegistry();
			putRegistry(context, registryType, projectRegistry);
		} else {
			if (namingRegistriesSubContext.containsKey(registryType)) {
				projectRegistry = (NamingRegistry)namingRegistriesSubContext.get(registryType);
			} else {
				projectRegistry = new ProjectNamingRegistry();
				namingRegistriesSubContext.put(registryType, projectRegistry);
			}
		}
		return projectRegistry;						

	}


	/**
	 * Create a unique key from the string
	 * @param string
	 * @return
	 */
	public static String getKeyFromString (String string) {
		int hashcode = string.hashCode();
		StringBuffer strBuf = new StringBuffer();
		String hashCodeString = Integer.toString(hashcode);
		if (hashcode < 0) {
			strBuf.append("0");
			strBuf.append(hashCodeString.substring(1));
		} else {
			strBuf.append(hashCodeString);
		}
		return strBuf.toString();		
	}	
	//defect 139867. generate name for the flow
	/** 
	*	returns the name of the flow.
	*/

	public static String getFlowName(Flow forType, TransformationRule rule){
		
		NameProvider nameProvider = NameProviderFactory.getNameProvider(forType);
		NamedElement sourceName = null;
		if(rule.getSource().get(0) instanceof PinSet){
			PinSet pinSet = (PinSet)rule.getSource().get(0);
			sourceName = (NamedElement)pinSet.eContainer();
		}else{
			sourceName =(NamedElement) rule.getSource().get(0);
		}
		//defect 140558
		String convertedName = nameProvider.getName(forType, sourceName, NamingUtil.findRegistry(rule));
		return convertedName;
		
		
	}
	
	/** 
	*	Get context for action.
	*/
	public static Object getContextContainer(Action source) {
		if (source == null) {
			return null;
		}
		
		Object contextContainer = null;
		if (BomUtils.isTopLevelProcess(source)) {
			contextContainer = ((StructuredActivityNode)source).getActivity();
		} else if (BomUtils.isReusableTask(source) || BomUtils.isReusableService(source)) {
			if(source instanceof CallBehaviorAction){
				contextContainer = ((CallBehaviorAction)source).getBehavior();
			} else if (source.getInStructuredNode() == null && source.eContainer() instanceof Activity){
				contextContainer = ((StructuredActivityNode)source).getActivity();
			}
		} else if (source.getInStructuredNode() != null) {	//local
			StructuredActivityNode san = source.getInStructuredNode();
			while (san.getInStructuredNode() != null) {		// get to the owning process
				san = san.getInStructuredNode();
			}
			contextContainer = san;
		} else if (source.getInStructuredNode() == null && source.eContainer() instanceof Activity) {
			contextContainer = (Activity)source.eContainer();
		} else {	//TODO any other containers?
			contextContainer = source;
		}
		return contextContainer;
	}

	
	public static String getNameForArtifact(String name) {
		if (name == null) {
			return null;
		}
		
		String registeredLastName = null;
		
		if (name.indexOf(ExportOperation.FILE_SEPARATOR) != -1) {
			registeredLastName = name.substring(name.lastIndexOf(ExportOperation.FILE_SEPARATOR) + 1);
		} else {
			registeredLastName = name;
		}
		
		return registeredLastName;
	}


	public static String getFolderNameForArtifact(String name) {
		if (name == null) {
			return null;
		}
		
		String folderName = null;
		
		if (name.indexOf(ExportOperation.FILE_SEPARATOR) != -1) {
			String pathNoFile = name.substring(0, name.lastIndexOf(ExportOperation.FILE_SEPARATOR));
			if (pathNoFile.indexOf(ExportOperation.FILE_SEPARATOR) != -1) {
				folderName = pathNoFile.substring(pathNoFile.lastIndexOf(ExportOperation.FILE_SEPARATOR) + 1);
			} else {
				folderName = pathNoFile;
			}
		}
		
		return folderName;
	}


	/**
	 * Returns a String that is a valid name.
	 * It converts the name into a valid name and registers name that is also unique in the registry	 
	 * @param action Action
	 * @param context TransformationContext	 
	 * @param name String name set by user in TA
	 * @return String the converted String name
	 */				
	public static String convertWSDLNameForAction(Action action, TransformationContext context, String name) {	
		if (action == null) {
			return null;
		}
		
		Action referencedAction = null;
		if (action instanceof CallBehaviorAction &&
				((CallBehaviorAction)action).getBehavior() instanceof Activity) {
			referencedAction = ((Activity)((CallBehaviorAction)action).getBehavior()).getImplementation();
		} else {
			referencedAction = action;
		}

		NoHashNamingRegistry noHashReg = null;
		if (NamingUtil.USE_PER_PROJECT_UNIQUENESS) {
			String projectName = CrossProjectReferenceUtil.getProjectNameForBOMElement(referencedAction);
			noHashReg = ((ProjectNamingRegistry)NamingUtil.getProjectNamingRegistry(context, NamingUtil.PROJECT_NAMING_REGISTRY)).getProjectRegistry(projectName);
		} else {
			noHashReg = ((ProjectNamingRegistry)NamingUtil.getProjectNamingRegistry(context, NamingUtil.PROJECT_NAMING_REGISTRY)).getProjectRegistry(NamingUtil.GLOBAL_XPROJECT_UNIQUENESS);
		}
		if(noHashReg.actionExistInWSDLRegistry(referencedAction)) {
			String registeredName = noHashReg.retrieveNameFromWSDLRegistry(referencedAction);
			return registeredName;
		} else {	// element not registered yet
			String newName = null;
			if (name != null) {	// use TA name
				newName = JavaNCNameConverter.replaceInvalidName(name); 
			} else {
				newName = JavaNCNameConverter.replaceInvalidName(referencedAction.getName());
			}
			// for local task get the path from the containing process 
			Object contextContainer = NamingUtil.getContextContainer(referencedAction);
			String pathName = null;
			if (contextContainer instanceof StructuredActivityNode && 
					noHashReg.actionExistInGlobalRegistry(contextContainer)) {
				String containerPath = noHashReg.retrieveNameFromGlobalRegistry((StructuredActivityNode)contextContainer);
				if (containerPath.indexOf(ExportOperation.FILE_SEPARATOR) != -1) {
					pathName = containerPath.substring(0, containerPath.lastIndexOf(ExportOperation.FILE_SEPARATOR) + 1) + newName;
				} else {
					pathName = containerPath + ExportOperation.FILE_SEPARATOR + newName;
				}
			}
			if (pathName == null) { 
				pathName = SAUtil.createPackage(referencedAction, ExportOperation.FILE_SEPARATOR) + newName;
			}
			if (!noHashReg.isUniqueNameInWSDLRegistry(pathName)) {
				String genName = generateNewWSDLName(pathName, referencedAction, noHashReg);
				noHashReg.registerWSDL(referencedAction, genName);
				return genName;					
			} else {
				noHashReg.registerWSDL(referencedAction, pathName);				
			}						
			return pathName;
		}
	}


	/**
	 * Returns a String that is a valid name.
	 * It converts the name into a valid name and registers name that is also unique in the registry	 
	 * @param action Action
	 * @param context TransformationContext	 
	 * @param name String name set by user in TA
	 * @return String the converted String name
	 */				
	public static String convertGlobalWSDLNameForAction(Action action, TransformationContext context, String name) {	
		if (action == null) {
			return null;
		}
		
		Action referencedAction = null;
		if (action instanceof CallBehaviorAction &&
					((CallBehaviorAction)action).getBehavior() instanceof Activity) {
			referencedAction = ((Activity)((CallBehaviorAction)action).getBehavior()).getImplementation();
		} else {
			referencedAction = action;
		}
		NoHashNamingRegistry noHashReg = null;
		if (NamingUtil.USE_PER_PROJECT_UNIQUENESS) {
			String projectName = CrossProjectReferenceUtil.getProjectNameForBOMElement(referencedAction);
			noHashReg = ((ProjectNamingRegistry)NamingUtil.getProjectNamingRegistry(context, NamingUtil.PROJECT_NAMING_REGISTRY)).getProjectRegistry(projectName);
		} else {
			noHashReg = ((ProjectNamingRegistry)NamingUtil.getProjectNamingRegistry(context, NamingUtil.PROJECT_NAMING_REGISTRY)).getProjectRegistry(NamingUtil.GLOBAL_XPROJECT_UNIQUENESS);
		}
		
		if(noHashReg.actionExistInWSDLRegistry(referencedAction)) {
			String registeredName = noHashReg.retrieveNameFromWSDLRegistry(referencedAction);
			return registeredName;
		} else {	// element not registered yet
			String newName = null;
			if (name != null) {	// use TA name
				newName = JavaNCNameConverter.replaceInvalidName(name); 
			} else {
				newName = JavaNCNameConverter.replaceInvalidName(referencedAction.getName());
			}
			String pathName = SAUtil.createPackage(referencedAction, ExportOperation.FILE_SEPARATOR) + newName;
			if (!noHashReg.isUniqueNameInWSDLRegistry(pathName)) {
				String genName = generateNewGlobalWSDLName(pathName, referencedAction, noHashReg);
				noHashReg.registerWSDL(referencedAction, genName);
				return genName;					
			} else {
				noHashReg.registerWSDL(referencedAction, pathName);				
			}						
			return pathName;
		}
	}

	/**
	 * Returns a String that is a valid name.
	 * It converts the name into a valid name that is also unique in the registry	 
	 * @param action Action
	 * @param context TransformationContext	 
	 * @return String the converted String name
	 */				
	public static String convertNameForAction(Action action, TransformationContext context, String name) {	
		if (action == null) {
			return null;
		}
		
		Action referencedAction = null;
		if (action instanceof CallBehaviorAction &&
				((CallBehaviorAction)action).getBehavior() instanceof Activity) {
			referencedAction = ((Activity)((CallBehaviorAction)action).getBehavior()).getImplementation();
		} else {
			referencedAction = action;
		}

		NoHashNamingRegistry noHashReg = null;
		if (NamingUtil.USE_PER_PROJECT_UNIQUENESS) {
			String projectName = CrossProjectReferenceUtil.getProjectNameForBOMElement(referencedAction);
			noHashReg = ((ProjectNamingRegistry)NamingUtil.getProjectNamingRegistry(context, NamingUtil.PROJECT_NAMING_REGISTRY)).getProjectRegistry(projectName);
		} else {
			noHashReg = ((ProjectNamingRegistry)NamingUtil.getProjectNamingRegistry(context, NamingUtil.PROJECT_NAMING_REGISTRY)).getProjectRegistry(NamingUtil.GLOBAL_XPROJECT_UNIQUENESS);
		}
		
		// Replace any invalid characters in the sourceName
		String newName = null;
		if (name != null) {
			newName = JavaNCNameConverter.replaceInvalidName(name);
		} else {
			newName = JavaNCNameConverter.replaceInvalidName(referencedAction.getName());
		}
		//defect 87022
		newName = includeTemplateVersionNumber(action, newName);

		// global element
		if (action instanceof CallBehaviorAction ||
				(action.getInStructuredNode() == null && action.eContainer() instanceof Activity)) {		
			if(noHashReg.actionExistInGlobalRegistry(referencedAction)) {
				String registeredName = noHashReg.retrieveNameFromGlobalRegistry(referencedAction);
				return registeredName;
			} else {	// global element not registered yet
				String pathName = SAUtil.createPackage(referencedAction, ExportOperation.FILE_SEPARATOR) + newName;
				if (!noHashReg.isUniqueName(referencedAction, pathName)) {
					String genName = generateGlobalName(pathName, referencedAction, noHashReg);
					noHashReg.register(referencedAction, genName);
					// register this action as its own container (case when process and local task have the same names)
					if (action instanceof StructuredActivityNode) {
						noHashReg.register(action, action, pathName);
					}
					return genName;					
				} else {
					noHashReg.register(referencedAction, pathName);
					// register this action as its own container (case when process and local task have the same names)
					if (action instanceof StructuredActivityNode) {
						noHashReg.register(action, action, pathName);
					}
				}
				return pathName;
			}
		} else {	// local element - use Local registry
			Object contextContainer = NamingUtil.getContextContainer(referencedAction);
			if (noHashReg.actionExistInLocalRegistry(referencedAction, contextContainer)) {
				String registeredName = noHashReg.retrieveNameFromLocalRegistry(referencedAction, contextContainer);
				return registeredName;
			} else {	// Action first time in registry
				String pathName = null;
				if(contextContainer instanceof StructuredActivityNode &&
						noHashReg.actionExistInGlobalRegistry(contextContainer)) {
					String containerPath = noHashReg.retrieveNameFromGlobalRegistry((StructuredActivityNode)contextContainer);
					pathName = null;
					if (containerPath.indexOf(ExportOperation.FILE_SEPARATOR) != -1) {
						pathName = containerPath.substring(0, containerPath.lastIndexOf(ExportOperation.FILE_SEPARATOR) + 1) + newName;
					} else {
						pathName = containerPath + ExportOperation.FILE_SEPARATOR + newName;;
					}
				} else  {
					pathName = SAUtil.createPackage(referencedAction, ExportOperation.FILE_SEPARATOR) + newName;
				}
				if (!noHashReg.isUniqueName(referencedAction, contextContainer, pathName)){	// checks for name
					String genName = generateNewName(pathName, referencedAction, contextContainer, noHashReg);
					noHashReg.register(referencedAction, contextContainer, genName);
					return genName;
				} else {
					noHashReg.register(referencedAction, contextContainer, pathName);
					return pathName;
				}
			}
		}		
	}

	
	/**
	 * Returns a String that is a valid name.
	 * It converts the name into a valid name that is also unique in the registry	 
	 * @param action Action
	 * @param context TransformationContext	 
	 * @return String the converted String name
	 */				
	public static String convertNameForAction(Action action, TransformationContext context) {	
		if (action == null) {
			return null;
		}
		
		Action referencedAction = null;
		if (action instanceof CallBehaviorAction &&
				((CallBehaviorAction)action).getBehavior() instanceof Activity) {
			if(BomUtils.isReusableBusinessRuleTask((CallBehaviorAction)action)||BomUtils.isReusableHumanTask((CallBehaviorAction)action)||BomUtils.isProcess((CallBehaviorAction)action)){
				referencedAction = ((Activity)((CallBehaviorAction)action).getBehavior()).getImplementation();
			}else{
				referencedAction = action;
			}
		} else {
			referencedAction = action;
		}

		NoHashNamingRegistry noHashReg = null;
		if (NamingUtil.USE_PER_PROJECT_UNIQUENESS) {
			String projectName = CrossProjectReferenceUtil.getProjectNameForBOMElement(referencedAction);
			noHashReg = ((ProjectNamingRegistry)NamingUtil.getProjectNamingRegistry(context, NamingUtil.PROJECT_NAMING_REGISTRY)).getProjectRegistry(projectName);
		} else {
			noHashReg = ((ProjectNamingRegistry)NamingUtil.getProjectNamingRegistry(context, NamingUtil.PROJECT_NAMING_REGISTRY)).getProjectRegistry(NamingUtil.GLOBAL_XPROJECT_UNIQUENESS);
		}
		// Replace any invalid characters in the sourceName
		String newName = null; 
			
		if (BomUtils.isServiceOperation(referencedAction)) {
			if (referencedAction.eContainer() instanceof Activity && ((Activity)referencedAction.eContainer()).getOwningPackage() instanceof ServiceInterface) {
				newName = JavaNCNameConverter.replaceInvalidName(((ServiceInterface)((Activity)referencedAction.eContainer()).getOwningPackage()).getName());
			} else {
				newName = JavaNCNameConverter.replaceInvalidName(referencedAction.getName());	//should not happen
			}
		} else {
			newName = JavaNCNameConverter.replaceInvalidName(referencedAction.getName());
		}
		
		newName = includeTemplateVersionNumber(action, newName);

		// global element
		if (action instanceof CallBehaviorAction ||
				(action.getInStructuredNode() == null && action.eContainer() instanceof Activity)) {		
			if(noHashReg.actionExistInGlobalRegistry(referencedAction)) {
				String registeredName = noHashReg.retrieveNameFromGlobalRegistry(referencedAction);
				return registeredName;
			} else {	// global element not registered yet
				String pathName = SAUtil.createPackage(referencedAction, ExportOperation.FILE_SEPARATOR) + newName;
				if (!noHashReg.isUniqueName(referencedAction, pathName)) {
					String genName = generateGlobalName(pathName, referencedAction, noHashReg);
					noHashReg.register(referencedAction, genName);
					// register this action as its own container (case when process and local task have the same names)
					if (action instanceof StructuredActivityNode) {
						noHashReg.register(action, action, pathName);
					}
					return genName;					
				} else {
					noHashReg.register(referencedAction, pathName);
					// register this action as its own container (case when process and local task have the same names)
					if (action instanceof StructuredActivityNode) {
						noHashReg.register(action, action, pathName);
					}
				}
				return pathName;
			}
		} else {	// local element - use Local registry
			Object contextContainer = NamingUtil.getContextContainer(referencedAction);
			if (noHashReg.actionExistInLocalRegistry(referencedAction, contextContainer)) {
				String registeredName = noHashReg.retrieveNameFromLocalRegistry(referencedAction, contextContainer);
				return registeredName;
			} else {	// Action first time in registry
				String pathName = null;
				if(contextContainer instanceof StructuredActivityNode &&
						noHashReg.actionExistInGlobalRegistry(contextContainer)) {
					String containerPath = noHashReg.retrieveNameFromGlobalRegistry((StructuredActivityNode)contextContainer);
					pathName = null;
					if (containerPath.indexOf(ExportOperation.FILE_SEPARATOR) != -1) {
						pathName = containerPath.substring(0, containerPath.lastIndexOf(ExportOperation.FILE_SEPARATOR) + 1) + newName;
					} else {
						pathName = containerPath + ExportOperation.FILE_SEPARATOR + newName;;
					}
				} else  {
					pathName = SAUtil.createPackage(referencedAction, ExportOperation.FILE_SEPARATOR) + newName;
				}
				if (!noHashReg.isUniqueName(referencedAction, contextContainer, pathName)){	// checks for name
					String genName = generateNewName(pathName, referencedAction, contextContainer, noHashReg);
					noHashReg.register(referencedAction, contextContainer, genName);
					return genName;
				} else {
					noHashReg.register(referencedAction, contextContainer, pathName);
					return pathName;
				}
			}
		}		
	}
	/**
	 * If there is template version specified in the context, should append the version number to the 
	 * generated default value. 
	 * Note: the version number will not be appended to the user specified value (through technical attributes).
	 * @param action
	 * @param newName
	 * @return
	 */
	public static String includeTemplateVersionNumber(Action action, String newName) {
		if(action.getInStructuredNode() == null && action.eContainer() instanceof Activity) {
			String blmID = ((Activity)action.eContainer()).getUid();
			String versionName = BpelOptionsUtil.getWPSTemplateSuffixForElement(blmID);
			if(versionName != null)	newName = newName + versionName;
		}else if(action instanceof CallBehaviorAction){
			Activity myActivity = ((Activity)((CallBehaviorAction)action).getBehavior()); 
			String blmID = myActivity.getUid();
			String versionName = BpelOptionsUtil.getWPSTemplateSuffixForElement(blmID);
			if(versionName != null)	newName = newName + versionName;
		}
		return newName;
	}


	/**
	 * Returns a String that is a valid name.
	 * It converts the name into a valid name that is also unique in the registry	 
	 * @param action Action
	 * @param context TransformationContext	 
	 * @return String the converted String name
	 */				
	public static String convertNameForImport (Action action, String attributeImportName, TransformationContext context) {	
		if (action == null) {
			return null;
		}
		NoHashNamingRegistry noHashReg = null;
		if (NamingUtil.USE_PER_PROJECT_UNIQUENESS) {
			String projectName = CrossProjectReferenceUtil.getProjectNameForBOMElement(action);
			noHashReg = ((ProjectNamingRegistry)NamingUtil.getProjectNamingRegistry(context, NamingUtil.PROJECT_NAMING_REGISTRY)).getProjectRegistry(projectName);
		} else {
			noHashReg = ((ProjectNamingRegistry)NamingUtil.getProjectNamingRegistry(context, NamingUtil.PROJECT_NAMING_REGISTRY)).getProjectRegistry(NamingUtil.GLOBAL_XPROJECT_UNIQUENESS);
		}
		Action referencedAction = null;
		if (action instanceof CallBehaviorAction && 
				((CallBehaviorAction)action).getBehavior() instanceof Activity) {
		   if(BomUtils.isReusableBusinessRuleTask((CallBehaviorAction)action)||BomUtils.isReusableHumanTask((CallBehaviorAction)action)||BomUtils.isProcess((CallBehaviorAction)action)){
			referencedAction = ((Activity)((CallBehaviorAction)action).getBehavior()).getImplementation();
		   }else{
			   referencedAction = action;
		   }
		} else {
			referencedAction = action;
		}
		
		String importName = noHashReg.retrieveNameFromGlobalRegistry(referencedAction);
		if (importName != null) {	// return registered name
			return importName;
		}
		
		String newImportName = null;
		String newName = null;
		if (attributeImportName == null) {
			newName = JavaNCNameConverter.replaceInvalidName(referencedAction.getName());
		} else {
			newName = JavaNCNameConverter.replaceInvalidName(attributeImportName);
		}
		
		Object contextContainer = NamingUtil.getContextContainer(referencedAction);
		if(contextContainer instanceof StructuredActivityNode &&
				noHashReg.actionExistInGlobalRegistry(contextContainer)) {
			String containerPath = noHashReg.retrieveNameFromGlobalRegistry((StructuredActivityNode)contextContainer);
			if (containerPath.indexOf(ExportOperation.FILE_SEPARATOR) != -1) {
				newImportName = containerPath.substring(0, containerPath.lastIndexOf(ExportOperation.FILE_SEPARATOR) + 1) + newName;
			} else {
				newImportName = containerPath + ExportOperation.FILE_SEPARATOR + newName;;
			}
		} else  {
			newImportName = SAUtil.createPackage(referencedAction, ExportOperation.FILE_SEPARATOR)+ newName;
		}
	
		if (!noHashReg.isUniqueName(referencedAction, newImportName)) {
			String genName = null;
			if (action instanceof CallBehaviorAction) {
				genName = generateGlobalName(newImportName, referencedAction, noHashReg);
			} else {
				genName = generateNewImportName(newImportName, referencedAction, noHashReg);
			}
			noHashReg.register(referencedAction, genName);
			// register in local registry to avoid name conflicts between export, import, and components
			noHashReg.register(referencedAction, contextContainer, genName);
			return genName;
		} else {
			noHashReg.register(referencedAction, newImportName);				
			// register in local registry to avoid name conflicts between export, import, and components
			noHashReg.register(referencedAction, contextContainer, newImportName);
		}
		
		return newImportName;
	}

	
	/**
	 * Returns a String that is a valid name.
	 * It converts the name into a valid name that is also unique in the registry	 
	 * @param action Action
	 * @param context TransformationContext	 
	 * @return String the converted String name
	 */				
	public static String convertNameForImport (Action action, TransformationContext context) {	
		if (action == null) {
			return null;
		}
		
		NoHashNamingRegistry noHashReg = null;
		if (NamingUtil.USE_PER_PROJECT_UNIQUENESS) {
			String projectName = CrossProjectReferenceUtil.getProjectNameForBOMElement(action);
			noHashReg = ((ProjectNamingRegistry)NamingUtil.getProjectNamingRegistry(context, NamingUtil.PROJECT_NAMING_REGISTRY)).getProjectRegistry(projectName);
		} else {
			noHashReg = ((ProjectNamingRegistry)NamingUtil.getProjectNamingRegistry(context, NamingUtil.PROJECT_NAMING_REGISTRY)).getProjectRegistry(NamingUtil.GLOBAL_XPROJECT_UNIQUENESS);
		}
		Action referencedAction = null;
		if (action instanceof CallBehaviorAction && 
				((CallBehaviorAction)action).getBehavior() instanceof Activity) {
			referencedAction = ((Activity)((CallBehaviorAction)action).getBehavior()).getImplementation();
		} else {
			referencedAction = action;
		}
		
		String importName = noHashReg.retrieveNameFromGlobalRegistry(referencedAction);
		if (importName != null) {	// return registered name
			return importName;
		}
		
		String newImportName = null;
		String newName = JavaNCNameConverter.replaceInvalidName(referencedAction.getName());
		
		Object contextContainer = NamingUtil.getContextContainer(referencedAction);
		if(contextContainer instanceof StructuredActivityNode &&
				noHashReg.actionExistInGlobalRegistry(contextContainer)) {
			String containerPath = noHashReg.retrieveNameFromGlobalRegistry((StructuredActivityNode)contextContainer);
			if (containerPath.indexOf(ExportOperation.FILE_SEPARATOR) != -1) {
				newImportName = containerPath.substring(0, containerPath.lastIndexOf(ExportOperation.FILE_SEPARATOR) + 1) + newName;
			} else {
				newImportName = containerPath + ExportOperation.FILE_SEPARATOR + newName;;
			}
		} else  {
			newImportName = SAUtil.createPackage(referencedAction, ExportOperation.FILE_SEPARATOR)+ newName;
		}
	
		newImportName = includeTemplateVersionNumber(action, newImportName);

		if (!noHashReg.isUniqueName(referencedAction, newImportName)) {
			String genName = null;
			if (action instanceof CallBehaviorAction) {
				genName = generateGlobalName(newImportName, referencedAction, noHashReg);
			} else {
				genName = generateNewName(newImportName, referencedAction, noHashReg);
			}
			noHashReg.register(referencedAction, genName);
			return genName;
		} else {
			noHashReg.register(referencedAction, newImportName);				
		}
		
		return newImportName;
	}

	
	/**
	 * Returns a new generated name that is unique in the registry 
	 * @param duplicatedName  the duplicated name that needs to be replaced
	 * @param registry  the namingRegistry
	 * @param obj the EObject construct
	 * @return String the newly generated string name
	 */	
	private static String generateGlobalName(String duplicatedName, Action action, NoHashNamingRegistry registry) {
		//instead of making the path/folderForComponent/fileNameForComponent2 create
		//		path/folderForComponent2/fileNameForComponent (for global elements only)
		if (duplicatedName == null || registry == null) {
			return null;
		}
		
		String fileName = null;
		String folderName = "";
		String path = "";
		
		if (duplicatedName.indexOf(ExportOperation.FILE_SEPARATOR) != -1) {
			fileName = duplicatedName.substring(duplicatedName.lastIndexOf(ExportOperation.FILE_SEPARATOR) + 1);
			String pathNoFile = duplicatedName.substring(0, duplicatedName.lastIndexOf(ExportOperation.FILE_SEPARATOR));
			if (pathNoFile.indexOf(ExportOperation.FILE_SEPARATOR) != -1) {
				folderName = pathNoFile.substring(pathNoFile.lastIndexOf(ExportOperation.FILE_SEPARATOR) + 1);
				path = pathNoFile.substring(0, pathNoFile.lastIndexOf(ExportOperation.FILE_SEPARATOR) + 1);   // include file separator
			} else {
				folderName = pathNoFile;
			}
		} else {
			fileName = duplicatedName;
		}
		
		int count = 2;
		String newFolderName =  "";
		if (!folderName.equals("")) {
			newFolderName =  folderName + "_" + count++;
		}			
		String newName = path + newFolderName + ExportOperation.FILE_SEPARATOR + fileName;
		// Keep trying to generate new name until name is unique
		while (!registry.isUniqueName(action, newName)) {
			if (folderName.equals("")) {
				newName = duplicatedName + "_" + count++;
			} else {
				newFolderName = folderName + "_" + count++;
				newName = path + newFolderName + ExportOperation.FILE_SEPARATOR + fileName;
			}
		}
		return newName;
	}
	
	
	/**
	 * Returns a new generated name that is unique in the registry 
	 * @param duplicatedName  the duplicated name that needs to be replaced
	 * @param registry  the namingRegistry
	 * @param obj the EObject construct
	 * @return String the newly generated string name
	 */	
	private static String generateNewGlobalWSDLName(String duplicatedName, Action action, NoHashNamingRegistry registry) {
		//instead of making the path/folderForComponent/fileNameForComponent2 create
		//		path/folderForComponent2/fileNameForComponent (for global elements only)
		if (duplicatedName == null || registry == null) {
			return null;
		}
		
		String fileName = null;
		String folderName = null;
		String path = "";
		
		if (duplicatedName.indexOf(ExportOperation.FILE_SEPARATOR) != -1) {
			fileName = duplicatedName.substring(duplicatedName.lastIndexOf(ExportOperation.FILE_SEPARATOR) + 1);
			String pathNoFile = duplicatedName.substring(0, duplicatedName.lastIndexOf(ExportOperation.FILE_SEPARATOR));
			if (pathNoFile.indexOf(ExportOperation.FILE_SEPARATOR) != -1) {
				folderName = pathNoFile.substring(pathNoFile.lastIndexOf(ExportOperation.FILE_SEPARATOR) + 1);
				path = pathNoFile.substring(0, pathNoFile.lastIndexOf(ExportOperation.FILE_SEPARATOR) + 1);   // include file separator
			} else {
				folderName = pathNoFile;
			}
		} else {
			fileName = duplicatedName;
		}
		
		int count = 2;
		String newFolderName = folderName + "_" + count++;
		String newName = path + newFolderName + ExportOperation.FILE_SEPARATOR + fileName;
		// Keep trying to generate new name until name is unique
		while (!registry.isUniqueNameInWSDLRegistry(newName)) {
			if (folderName == null) {
				newName = duplicatedName + "_" + count++;
			} else {
				newFolderName = folderName + "_" + count++;
				newName = path + newFolderName + ExportOperation.FILE_SEPARATOR + fileName;
			}
		}
		return newName;
	}

	
	/**
	 * Returns a new generated name that is unique in the registry 
	 * @param duplicatedName  the duplicated name that needs to be replaced
	 * @param registry  the namingRegistry
	 * @param obj the EObject construct
	 * @return String the newly generated string name
	 */	
	private static String generateNewWSDLName(String duplicatedName, Action action, NoHashNamingRegistry registry)
	{
		int count = 2;
		String newName = duplicatedName + "_" + count++;
		// Keep trying to generate new name until name is unique
		while (!registry.isUniqueNameInWSDLRegistry(newName)) {
			newName = duplicatedName + "_" + count++;
		}
		return newName;
	}

	/**
	 * Returns a new generated name that is unique in the registry 
	 * @param duplicatedName  the duplicated name that needs to be replaced
	 * @param registry  the namingRegistry
	 * @param obj the EObject construct
	 * @return String the newly generated string name
	 */	
	private static String generateNewName(String duplicatedName, Object obj, NamingRegistry registry)
	{
		int count = 2;
		String newName = duplicatedName + "_" + count++;
		// Keep trying to generate new name until name is unique
		while (!registry.isUniqueName(obj, newName))
		{
			newName = duplicatedName + "_" + count++;
		}
		return newName;
	}
	

	/**
	 * Returns a new generated name that is unique in the registry 
	 * @param duplicatedName  the duplicated name that needs to be replaced
	 * @param registry  the namingRegistry
	 * @param obj the EObject construct
	 * @return String the newly generated string name
	 */	
	private static String generateNewImportName(String duplicatedName, Object obj, NamingRegistry registry)
	{
		String newName = null;
		if (duplicatedName != null &&
				duplicatedName.length() > 2 &&
				duplicatedName.indexOf("_") == duplicatedName.length() - 2) {
			int count = 2;
			try {
				NumberFormat numberFormat = NumberFormat.getInstance();
				Number number = numberFormat.parse(duplicatedName.substring(duplicatedName.length() - 1));
				count = number.intValue();
				count++;
				newName = duplicatedName.substring(0, duplicatedName.length() - 2) + "_" + count++;
				// Keep trying to generate new name until name is unique
				while (!registry.isUniqueName(obj, newName))
				{
					newName = duplicatedName.substring(0, duplicatedName.length() - 2) + "_" + count++;
				}
			} catch (ParseException pe) {
				newName = duplicatedName + "_" + count++;
				// Keep trying to generate new name until name is unique
				while (!registry.isUniqueName(obj, newName))
				{
					newName = duplicatedName + "_" + count++;
				}				
			}
		} else {
			int count = 2;
			newName = duplicatedName + "_" + count++;
			// Keep trying to generate new name until name is unique
			while (!registry.isUniqueName(obj, newName))
			{
				newName = duplicatedName + "_" + count++;
			}
		}
		return newName;
	}
	

	/**
	 * Returns a new generated name that is unique in the registry 
	 * @param duplicatedName  the duplicated name that needs to be replaced
	 * @param registry  the namingRegistry
	 * @param obj the EObject construct
	 * @return String the newly generated string name
	 */	
	private static String generateNewName(String duplicatedName, Object obj, Object parentContext, NamingRegistry registry)
	{
		int count = 2;
		String newName = duplicatedName + "_" + count++;
		// Keep trying to generate new name until name is unique
		while (!registry.isUniqueName(obj, parentContext, newName))
		{
			newName = duplicatedName + "_" + count++;
		}
		return newName;
	}	
	
	
	public static void cleanNamingRegistries(TransformationContext context){
		// assert global naming registry
		NoHashNamingRegistry noHashReg = ((ProjectNamingRegistry)NamingUtil.getProjectNamingRegistry(context, NamingUtil.PROJECT_NAMING_REGISTRY)).getProjectRegistry(NamingUtil.GLOBAL_XPROJECT_UNIQUENESS);
		if(noHashReg != null){
			if (noHashReg.getGlobalContextRegistry() != null) 
				noHashReg.getGlobalContextRegistry().clear();
			if (noHashReg.getWSDLRegistry() != null)
				noHashReg.getWSDLRegistry().clear();
			if (noHashReg.getLocalContextRegistries() != null) {
				HashMap<Object, HashMap<Action, String>> map = noHashReg.getLocalContextRegistries();
				if (map != null) {
					map.clear();
				}
			}
		}
		
		BRNamingRegistry nReg = (BRNamingRegistry)getBRNamingRegistry(context, BR_NAMING_REGISTRY);
		if (nReg != null) {
			if (nReg.getRegistry() != null)
				nReg.getRegistry().clear();
			if (nReg.getBrGroupNameRegistry() != null)
				nReg.getBrGroupNameRegistry().clear();
			if (nReg.getBrSetNameRegistry() != null)
				nReg.getBrSetNameRegistry().clear();
			if (nReg.getBrTemplateNameRegistry() != null)
				nReg.getBrTemplateNameRegistry().clear();
		}
		
		
		ScdlNamingRegistry scdlReg = (ScdlNamingRegistry)getDefaultRegistry(context, DEFAULT_REGISTRY_SCDL);
		if (scdlReg != null && scdlReg.getRegistry() != null) {
			scdlReg.getRegistry().clear();
		}
	}

	
	/**
	 * Returns a new generated name that is unique in the registry 
	 * @param duplicatedName  the duplicated name that needs to be replaced
 	 * @param rti RecurringTimeIntervals
	 * @param registry  the namingRegistry
	 * @return String the newly generated string name
	 */	
	public static String generateNewTimetableName(String duplicatedName, NoHashNamingRegistry registry) {
		//instead of making the path/folderForComponent/fileNameForComponent2 create
		//		path/folderForComponent2/fileNameForComponent (for global elements only)
		if (duplicatedName == null || registry == null) {
			return null;
		}
		
		String fileName = null;
		String folderName = null;
		String path = "";
		
		if (duplicatedName.indexOf(ExportOperation.FILE_SEPARATOR) != -1) {
			fileName = duplicatedName.substring(duplicatedName.lastIndexOf(ExportOperation.FILE_SEPARATOR) + 1);
			String pathNoFile = duplicatedName.substring(0, duplicatedName.lastIndexOf(ExportOperation.FILE_SEPARATOR));
			if (pathNoFile.indexOf(ExportOperation.FILE_SEPARATOR) != -1) {
				folderName = pathNoFile.substring(pathNoFile.lastIndexOf(ExportOperation.FILE_SEPARATOR) + 1);
				path = pathNoFile.substring(0, pathNoFile.lastIndexOf(ExportOperation.FILE_SEPARATOR) + 1);   // include file separator
			} else {
				folderName = pathNoFile;
			}
		} else {
			fileName = duplicatedName;
		}
		
		int count = 2;
		String newFolderName = folderName + "_" + count++;
		String newName = path + newFolderName + ExportOperation.FILE_SEPARATOR + fileName;
		// Keep trying to generate new name until name is unique
		while (!registry.isUniqueNameInTimetableRegistry(newName)) {
			if (folderName == null) {
				newName = duplicatedName + "_" + count++;
			} else {
				newFolderName = folderName + "_" + count++;
				newName = path + newFolderName + ExportOperation.FILE_SEPARATOR + fileName;
			}
		}
		return newName;
	}

	// return the registered name
	public static String registerNameForTimetable(RecurringTimeIntervals rti, TransformationContext context) {
		if (rti == null) {
			return null;
		}
		
		NoHashNamingRegistry noHashReg = null;
		if (NamingUtil.USE_PER_PROJECT_UNIQUENESS) {
			String projectName = CrossProjectReferenceUtil.getProjectNameForBOMElement(rti);
			noHashReg = ((ProjectNamingRegistry)NamingUtil.getProjectNamingRegistry(context, NamingUtil.PROJECT_NAMING_REGISTRY)).getProjectRegistry(projectName);
		} else {
			noHashReg = ((ProjectNamingRegistry)NamingUtil.getProjectNamingRegistry(context, NamingUtil.PROJECT_NAMING_REGISTRY)).getProjectRegistry(NamingUtil.GLOBAL_XPROJECT_UNIQUENESS);
		}
		
		String registeredName = null;
		if(noHashReg.timetableExistInTimetableRegistry(rti)) {
			registeredName = noHashReg.retrieveNameFromTimetableRegistry(rti);
		} else {	// element not registered yet
			String name = JavaNCNameConverter.replaceInvalidName(rti.getName());
			String pathName = SAUtil.createPackage(rti, ExportOperation.FILE_SEPARATOR) + name;
			if (!noHashReg.isUniqueNameInTimetableRegistry(pathName)) {
				String genName = generateNewTimetableName(pathName, noHashReg);
				noHashReg.registerTimetable(rti, genName);
				registeredName = genName;					
			} else {
				noHashReg.registerTimetable(rti, pathName);
				registeredName = pathName;
			}						
		}
		return registeredName; 
	}	

}
